"""Inference Task of OTX Classification."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os
from typing import Optional

import numpy as np
from mmcv.utils import ConfigDict

from otx.algorithms.classification.adapters.mmcls.utils.builder import build_classifier
from otx.algorithms.classification.adapters.mmcls.utils.config_utils import (
    patch_datasets,
    patch_evaluation,
)
from otx.algorithms.classification.configs import ClassificationConfig
from otx.algorithms.classification.utils import (
    get_multihead_class_info as get_hierarchical_info,
)
from otx.algorithms.common.adapters.mmcv.utils import (
    patch_data_pipeline,
    patch_default_config,
    patch_runner,
)
from otx.algorithms.common.configs import TrainType
from otx.algorithms.common.tasks import BaseTask
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.metadata import FloatMetadata, FloatType
from otx.api.entities.model import (  # ModelStatus
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.explain_interface import IExplainTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.dataset_utils import add_saliency_maps_to_dataset_item
from otx.api.utils.labels_utils import get_empty_label
from otx.mpa import MPAConstants
from otx.mpa.utils.config_utils import MPAConfig
from otx.mpa.utils.logger import get_logger

# pylint: disable=invalid-name

logger = get_logger()

TASK_CONFIG = ClassificationConfig
RECIPE_TRAIN_TYPE = {
    TrainType.SEMISUPERVISED: "semisl.yaml",
    TrainType.INCREMENTAL: "incremental.yaml",
    TrainType.SELFSUPERVISED: "selfsl.yaml",
}


class ClassificationInferenceTask(
    BaseTask, IInferenceTask, IExportTask, IEvaluationTask, IExplainTask, IUnload
):  # pylint: disable=too-many-instance-attributes
    """Inference Task Implementation of OTX Classification."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment, **kwargs):
        self._should_stop = False
        super().__init__(TASK_CONFIG, task_environment, **kwargs)

        self._task_environment = task_environment
        if len(task_environment.get_labels(False)) == 1:
            self._labels = task_environment.get_labels(include_empty=True)
        else:
            self._labels = task_environment.get_labels(include_empty=False)
        self._empty_label = get_empty_label(task_environment.label_schema)

        self._multilabel = False
        self._hierarchical = False
        self._selfsl = False

        self._multilabel = len(task_environment.label_schema.get_groups(False)) > 1 and len(
            task_environment.label_schema.get_groups(False)
        ) == len(
            task_environment.get_labels(include_empty=False)
        )  # noqa:E127
        if self._multilabel:
            logger.info("Classification mode: multilabel")

        self._hierarchical_info = None
        if not self._multilabel and len(task_environment.label_schema.get_groups(False)) > 1:
            logger.info("Classification mode: hierarchical")
            self._hierarchical = True
            self._hierarchical_info = get_hierarchical_info(task_environment.label_schema)
        if not self._multilabel and not self._hierarchical:
            logger.info("Classification mode: multiclass")

        if self._hyperparams.algo_backend.train_type == TrainType.SELFSUPERVISED:
            self._selfsl = True

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self,
        dataset: DatasetEntity,
        inference_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main infer function of OTX Classification."""

        logger.info("called infer()")
        stage_module = "ClsInferrer"
        self._data_cfg = self._init_test_data_cfg(dataset)
        dataset = dataset.with_empty_annotations()

        dump_features = True
        dump_saliency_map = not inference_parameters.is_evaluation if inference_parameters else True

        results = self._run_task(
            stage_module,
            mode="eval",
            dataset=dataset,
            dump_features=dump_features,
            dump_saliency_map=dump_saliency_map,
        )
        logger.debug(f"result of run_task {stage_module} module = {results}")
        predictions = results["outputs"]
        prediction_results = zip(
            predictions["eval_predictions"],
            predictions["feature_vectors"],
            predictions["saliency_maps"],
        )

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress  # type: ignore
            process_saliency_maps = inference_parameters.process_saliency_maps
            explain_predicted_classes = inference_parameters.explain_predicted_classes

        self._add_predictions_to_dataset(
            prediction_results, dataset, update_progress_callback, process_saliency_maps, explain_predicted_classes
        )
        return dataset

    def explain(
        self,
        dataset: DatasetEntity,
        explain_parameters: Optional[InferenceParameters] = None,
    ) -> DatasetEntity:
        """Main explain function of OTX Classification Task."""
        logger.info("called explain()")
        stage_module = "ClsExplainer"
        self._data_cfg = self._init_test_data_cfg(dataset)
        dataset = dataset.with_empty_annotations()

        results = self._run_task(
            stage_module,
            mode="train",
            dataset=dataset,
            explainer=explain_parameters.explainer if explain_parameters else None,
        )
        logger.debug(f"result of run_task {stage_module} module = {results}")
        predictions = results["outputs"]["eval_predictions"]
        saliency_maps = results["outputs"]["saliency_maps"]

        update_progress_callback = default_progress_callback
        process_saliency_maps = False
        explain_predicted_classes = True
        if explain_parameters is not None:
            update_progress_callback = explain_parameters.update_progress  # type: ignore
            process_saliency_maps = explain_parameters.process_saliency_maps
            explain_predicted_classes = explain_parameters.explain_predicted_classes

        self._add_explanations_to_dataset(
            predictions,
            saliency_maps,
            dataset,
            update_progress_callback,
            process_saliency_maps,
            explain_predicted_classes,
        )
        logger.info("Explain completed")
        return dataset

    @check_input_parameters_type()
    def evaluate(
        self,
        output_resultset: ResultSetEntity,
        evaluation_metric: Optional[str] = None,
    ):
        """Evaluate function of OTX Classification Task."""

        logger.info("called evaluate()")
        metric = MetricsHelper.compute_accuracy(output_resultset)
        logger.info(f"Accuracy after evaluation: {metric.accuracy.value}")
        output_resultset.performance = metric.get_performance()
        logger.info("Evaluation completed")

    def unload(self):
        """Unload function of OTX Classification Task."""
        logger.info("called unload()")
        self.cleanup()

    @check_input_parameters_type()
    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = False):
        """Export function of OTX Classification Task."""

        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO

        stage_module = "ClsExporter"
        results = self._run_task(stage_module, mode="train", export=True, dump_features=dump_features)
        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        if xml_file is None or bin_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        output_model.precision = [ModelPrecision.FP32]
        output_model.set_data(
            "label_schema.json",
            label_schema_to_bytes(self._task_environment.label_schema),
        )
        logger.info("Exporting completed")

    # pylint: disable=too-many-locals
    def _get_item_labels(self, prediction_item, pos_thr):
        item_labels = []

        if self._multilabel:
            if max(prediction_item) < pos_thr:
                logger.info("Confidence is smaller than pos_thr, empty_label will be appended to item_labels.")
                item_labels.append(ScoredLabel(self._empty_label, probability=1.0))
            else:
                for cls_idx, pred_item in enumerate(prediction_item):
                    if pred_item > pos_thr:
                        cls_label = ScoredLabel(self.labels[cls_idx], probability=float(pred_item))
                        item_labels.append(cls_label)

        elif self._hierarchical:
            for head_idx in range(self._hierarchical_info["num_multiclass_heads"]):
                logits_begin, logits_end = self._hierarchical_info["head_idx_to_logits_range"][head_idx]
                head_logits = prediction_item[logits_begin:logits_end]
                head_pred = np.argmax(head_logits)  # Assume logits already passed softmax
                label_str = self._hierarchical_info["all_groups"][head_idx][head_pred]
                otx_label = next(x for x in self._labels if x.name == label_str)
                item_labels.append(ScoredLabel(label=otx_label, probability=float(head_logits[head_pred])))

            if self._hierarchical_info["num_multilabel_classes"]:
                head_logits = prediction_item[self._hierarchical_info["num_single_label_classes"] :]
                for logit_idx, logit in enumerate(head_logits):
                    if logit > pos_thr:  # Assume logits already passed sigmoid
                        label_str_idx = self._hierarchical_info["num_multiclass_heads"] + logit_idx
                        label_str = self._hierarchical_info["all_groups"][label_str_idx][0]
                        otx_label = next(x for x in self._labels if x.name == label_str)
                        item_labels.append(ScoredLabel(label=otx_label, probability=float(logit)))
            item_labels = self._task_environment.label_schema.resolve_labels_probabilistic(item_labels)
            if not item_labels:
                logger.info("item_labels is empty.")
                item_labels.append(ScoredLabel(self._empty_label, probability=1.0))

        else:
            label_idx = prediction_item.argmax()
            cls_label = ScoredLabel(
                self._labels[label_idx],
                probability=float(prediction_item[label_idx]),
            )
            item_labels.append(cls_label)
        return item_labels

    # pylint: disable=too-many-branches, too-many-locals
    def _add_predictions_to_dataset(
        self,
        prediction_results,
        dataset,
        update_progress_callback,
        process_saliency_maps=False,
        explain_predicted_classes=True,
    ):
        """Loop over dataset again to assign predictions.Convert from MMClassification format to OTX format."""

        dataset_size = len(dataset)
        pos_thr = 0.5
        for i, (dataset_item, prediction_items) in enumerate(zip(dataset, prediction_results)):
            prediction_item, feature_vector, saliency_map = prediction_items
            if any(np.isnan(prediction_item)):
                logger.info("Nan in prediction_item.")

            item_labels = self._get_item_labels(prediction_item, pos_thr)

            dataset_item.append_labels(item_labels)

            probs = TensorEntity(name="probabilities", numpy=prediction_item.reshape(-1))
            dataset_item.append_metadata_item(probs, model=self._task_environment.model)

            top_idxs = np.argpartition(prediction_item, -2)[-2:]
            top_probs = prediction_item[top_idxs]
            active_score_media = FloatMetadata(
                name="active_score", value=top_probs[1] - top_probs[0], float_type=FloatType.ACTIVE_SCORE
            )
            dataset_item.append_metadata_item(active_score_media, model=self._task_environment.model)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if saliency_map is not None:
                add_saliency_maps_to_dataset_item(
                    dataset_item=dataset_item,
                    saliency_map=saliency_map,
                    model=self._task_environment.model,
                    labels=self._labels,
                    predicted_scored_labels=item_labels,
                    explain_predicted_classes=explain_predicted_classes,
                    process_saliency_maps=process_saliency_maps,
                )
            update_progress_callback(int(i / dataset_size * 100))

    def _add_explanations_to_dataset(
        self,
        predictions,
        saliency_maps,
        dataset,
        update_progress_callback,
        process_saliency_maps,
        explain_predicted_classes,
    ):
        """Loop over dataset again and assign saliency maps."""
        dataset_size = len(dataset)
        for i, (dataset_item, prediction_item, saliency_map) in enumerate(zip(dataset, predictions, saliency_maps)):
            item_labels = self._get_item_labels(prediction_item, pos_thr=0.5)
            add_saliency_maps_to_dataset_item(
                dataset_item=dataset_item,
                saliency_map=saliency_map,
                model=self._task_environment.model,
                labels=self._labels,
                predicted_scored_labels=item_labels,
                explain_predicted_classes=explain_predicted_classes,
                process_saliency_maps=process_saliency_maps,
            )
            update_progress_callback(int(i / dataset_size * 100))

    def _init_recipe_hparam(self) -> dict:
        params = self._hyperparams.learning_parameters
        warmup_iters = int(params.learning_rate_warmup_iters)
        if self._multilabel:
            # hack to use 1cycle policy
            lr_config = ConfigDict(max_lr=params.learning_rate, warmup=None)
        else:
            lr_config = (
                ConfigDict(warmup_iters=warmup_iters) if warmup_iters > 0 else ConfigDict(warmup_iters=0, warmup=None)
            )

        early_stop = False
        if self._recipe_cfg is not None:
            if params.enable_early_stopping and self._recipe_cfg.get("evaluation", None):
                early_stop = ConfigDict(
                    start=int(params.early_stop_start),
                    patience=int(params.early_stop_patience),
                    iteration_patience=int(params.early_stop_iteration_patience),
                )

        if self._recipe_cfg.runner.get("type").startswith("IterBasedRunner"):  # type: ignore
            runner = ConfigDict(max_iters=int(params.num_iters))
        else:
            runner = ConfigDict(max_epochs=int(params.num_iters))

        config = ConfigDict(
            optimizer=ConfigDict(lr=params.learning_rate),
            lr_config=lr_config,
            early_stop=early_stop,
            data=ConfigDict(
                samples_per_gpu=int(params.batch_size),
                workers_per_gpu=int(params.num_workers),
            ),
            runner=runner,
        )

        if self._train_type.value == "SEMISUPERVISED":
            unlabeled_config = ConfigDict(
                data=ConfigDict(
                    unlabeled_dataloader=ConfigDict(
                        samples_per_gpu=int(params.unlabeled_batch_size),
                        workers_per_gpu=int(params.num_workers),
                    )
                )
            )
            config.update(unlabeled_config)
        return config

    def _init_recipe(self):
        logger.info("called _init_recipe()")

        if self._multilabel:
            recipe_root = os.path.join(MPAConstants.RECIPES_PATH, "stages/classification/multilabel")
        else:
            recipe_root = os.path.join(MPAConstants.RECIPES_PATH, "stages/classification")

        logger.info(f"train type = {self._train_type}")

        if self._train_type in RECIPE_TRAIN_TYPE:
            # TODO: this condition will be simplified after adding support for multlabel and hierarchical to SupCon.
            if (
                self._train_type == TrainType.INCREMENTAL
                and not self._multilabel
                and not self._hierarchical
                and self._hyperparams.learning_parameters.enable_supcon
                and not self._model_dir.endswith("supcon")
            ):
                recipe = os.path.join(recipe_root, "supcon.yaml")
                self._model_dir = os.path.join(self._model_dir, "supcon")
            else:
                recipe = os.path.join(recipe_root, RECIPE_TRAIN_TYPE[self._train_type])
        else:
            raise NotImplementedError(f"Train type {self._train_type} is not implemented yet.")

        logger.info(f"train type = {self._train_type} - loading {recipe}")

        self._recipe_cfg = MPAConfig.fromfile(recipe)

        # FIXME[Soobee] : if train type is not in cfg, it raises an error in default INCREMENTAL mode.
        # During semi-implementation, this line should be fixed to -> self._recipe_cfg.train_type = train_type
        self._recipe_cfg.train_type = self._train_type.name

        options_for_patch_datasets = {"type": "OTXClsDataset", "empty_label": self._empty_label}
        options_for_patch_evaluation = {"task": "normal"}
        if self._multilabel:
            options_for_patch_datasets["type"] = "OTXMultilabelClsDataset"
            options_for_patch_evaluation["task"] = "multilabel"
        elif self._hierarchical:
            options_for_patch_datasets["type"] = "OTXHierarchicalClsDataset"
            options_for_patch_datasets["hierarchical_info"] = self._hierarchical_info
            options_for_patch_evaluation["task"] = "hierarchical"
        elif self._selfsl:
            options_for_patch_datasets["type"] = "SelfSLDataset"
        patch_default_config(self._recipe_cfg)
        patch_runner(self._recipe_cfg)
        patch_data_pipeline(self._recipe_cfg, self.data_pipeline_path)
        patch_datasets(
            self._recipe_cfg,
            self._task_type.domain,
            **options_for_patch_datasets,
        )  # for OTX compatibility
        patch_evaluation(self._recipe_cfg, **options_for_patch_evaluation)  # for OTX compatibility
        logger.info(f"initialized recipe = {recipe}")

    # TODO: make cfg_path loaded from custom model cfg file corresponding to train_type
    # model.py contains heads/classifier only for INCREMENTAL setting
    # error log : ValueError: Unexpected type of 'data_loader' parameter
    def _init_model_cfg(self):
        if self._multilabel:
            cfg_path = os.path.join(self._model_dir, "model_multilabel.py")
        elif self._hierarchical:
            cfg_path = os.path.join(self._model_dir, "model_hierarchical.py")
        else:
            cfg_path = os.path.join(self._model_dir, "model.py")
        cfg = MPAConfig.fromfile(cfg_path)

        cfg.model.multilabel = self._multilabel
        cfg.model.hierarchical = self._hierarchical
        if self._hierarchical:
            cfg.model.head.hierarchical_info = self._hierarchical_info
        return cfg

    def _init_test_data_cfg(self, dataset: DatasetEntity):
        data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )
        return data_cfg

    def _update_stage_module(self, stage_module):
        module_prefix = {TrainType.INCREMENTAL: "Incr", TrainType.SEMISUPERVISED: "SemiSL"}
        if self._train_type in module_prefix and stage_module in ["ClsTrainer", "ClsInferrer"]:
            stage_module = module_prefix[self._train_type] + stage_module
        return stage_module

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)
        options["model_builder"] = build_classifier



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn

from ..utils import AngularPWConv, normalize


class SegmentOutNormMixin(nn.Module):
    def __init__(self, *args, enable_out_seg=True, enable_out_norm=False, **kwargs):
        super().__init__(*args, **kwargs)

        self.enable_out_seg = enable_out_seg
        self.enable_out_norm = enable_out_norm

        if enable_out_seg:
            if enable_out_norm:
                self.conv_seg = AngularPWConv(self.channels, self.out_channels, clip_output=True)
        else:
            self.conv_seg = None

    def cls_seg(self, feat):
        """Classify each pixel."""
        if self.dropout is not None:
            feat = self.dropout(feat)
        if self.enable_out_norm:
            feat = normalize(feat, dim=1, p=2)
        if self.conv_seg is not None:
            return self.conv_seg(feat)
        else:
            return feat



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch.nn as nn
from mmdet.models.builder import NECKS
from mmdet.models.necks.fpn import FPN

from ...mmov_model import MMOVModel


@NECKS.register_module()
class MMOVFPN(FPN):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        init_weight: bool = False,
        verify_shape: bool = True,
        *args,
        **kwargs
    ):

        # dummy
        in_channels = [8 for _ in inputs["laterals"]]
        out_channels = 8
        relu_before_extra_convs = False
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            relu_before_extra_convs=relu_before_extra_convs * args,
            **kwargs
        )

        self.lateral_convs = nn.ModuleList()
        for input, output in zip(inputs["laterals"], outputs["laterals"]):
            self.lateral_convs.append(
                MMOVModel(
                    model_path_or_model,
                    weight_path,
                    inputs=input,
                    outputs=output,
                    remove_normalize=False,
                    merge_bn=False,
                    paired_bn=False,
                    init_weight=init_weight,
                    verify_shape=verify_shape,
                )
            )

        self.fpn_convs = nn.ModuleList()
        for input, output in zip(inputs["fpn"], outputs["fpn"]):
            if input and output:
                self.fpn_convs.append(
                    MMOVModel(
                        model_path_or_model,
                        weight_path,
                        inputs=input,
                        outputs=output,
                        remove_normalize=False,
                        merge_bn=True,
                        paired_bn=True,
                        init_weight=init_weight,
                        verify_shape=verify_shape,
                    )
                )
            else:
                self.fpn_convs.append(nn.Identity())

    def init_weights(self, pretrained=None):
        # TODO
        pass



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F

from .normalize import normalize


class AngularPWConv(nn.Module):
    def __init__(self, in_features, out_features, clip_output=False):
        super(AngularPWConv, self).__init__()

        self.in_features = in_features
        assert in_features > 0
        self.out_features = out_features
        assert out_features >= 2
        self.clip_output = clip_output

        self.weight = nn.Parameter(torch.Tensor(out_features, in_features))
        self.weight.data.normal_().renorm_(2, 1, 1e-5).mul_(1e5)

    def forward(self, x):
        weight = normalize(self.weight, dim=1, p=2).view(self.out_features, self.in_features, 1, 1)
        out = F.conv2d(x, weight)

        if self.clip_output and not torch.onnx.is_in_onnx_export():
            out = out.clamp(-1.0, 1.0)

        return out



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import numpy as np
import pytest
from openvino.model_zoo.model_api.adapters.openvino_adapter import OpenvinoAdapter

from otx.algorithms.segmentation.adapters.openvino.model_wrappers.blur import (
    BlurSegmentation,
    get_activation_map,
)
from otx.api.usecases.adapters.model_adapter import IDataSource
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class MockDataSource(IDataSource):
    def __init__(self):
        pass

    @property
    def data(self):
        return None


class MockBlurSegmentation(BlurSegmentation):
    def __init__(self):
        pass


class MockAdapter(OpenvinoAdapter):
    def __init__(self):
        pass


class TestBlurSegmentationParamsValidation:
    @e2e_pytest_unit
    def test_blur_segmentation_init_params_validation(self):
        """
        <b>Description:</b>
        Check BlurSegmentation object initialization parameters validation

        <b>Input data:</b>
        BlurSegmentation object initialization parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        BlurSegmentation object initialization parameter
        """
        model_adapter = MockAdapter()
        correct_values_dict = {
            "model_adapter": model_adapter,
        }
        unexpected_str = "unexpected string"
        unexpected_values = [
            # Unexpected string is specified as "model_adapter" parameter
            ("model_adapter", unexpected_str),
            # Unexpected string is specified as "configuration" parameter
            ("configuration", unexpected_str),
            # Unexpected string is specified as "preload" parameter
            ("preload", unexpected_str),
        ]

        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=BlurSegmentation,
        )

    @e2e_pytest_unit
    def test_blur_segmentation_postprocess_params_validation(self):
        """
        <b>Description:</b>
        Check BlurSegmentation object "postprocess" method input parameters validation

        <b>Input data:</b>
        BlurSegmentation object, "postprocess" method unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "postprocess" method
        """
        random_array = np.random.randint(low=0, high=255, size=(10, 16, 3))
        blur_segmentation = MockBlurSegmentation()
        correct_values_dict = {
            "outputs": {"output_1": random_array},
            "metadata": {"metadata_1": "some metadata"},
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "outputs" parameter
            ("outputs", unexpected_int),
            # Unexpected integer is specified as "outputs" dictionary key
            ("outputs", {unexpected_int: random_array}),
            # Unexpected integer is specified as "outputs" dictionary value
            ("outputs", {"output_1": unexpected_int}),
            # Unexpected integer is specified as "metadata" parameter
            ("metadata", unexpected_int),
            # Unexpected integer is specified as "metadata" dictionary key
            ("metadata", {unexpected_int: "some metadata"}),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=blur_segmentation.postprocess,
        )


class TestBlurUtilsParamsValidation:
    @e2e_pytest_unit
    def test_get_activation_map_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_activation_map" function input parameters validation

        <b>Input data:</b>
        "features" unexpected type object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "get_activation_map" function
        """
        with pytest.raises(ValueError):
            get_activation_map(features=None)  # type: ignore



"""Data loaders for OpenVINO action recognition models."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from copy import deepcopy
from typing import Dict, List

import numpy as np
from compression.api import DataLoader

from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity, DatasetItemEntity


def get_ovdataloader(dataset: DatasetEntity, task_type: str, clip_len: int, width: int, height: int) -> DataLoader:
    """Find proper dataloader for dataset and task type.

    If dataset has only a single video, this returns DataLoader for online demo
    If dataset has multiple videos, this return DataLoader for academia evaluation
    """
    if _is_multi_video(dataset):
        if task_type == "ACTION_CLASSIFICATION":
            return ActionOVClsDataLoader(dataset, clip_len, width, height)
        if task_type == "ACTION_DETECTION":
            return ActionOVDetDataLoader(dataset, clip_len, width, height)
        raise NotImplementedError(f"{task_type} is not supported from action task")
    return ActionOVDemoDataLoader(dataset, task_type, clip_len, width, height)


def _is_multi_video(dataset: DatasetEntity) -> bool:
    """Check dataset has multiple videos."""
    _video_id = dataset[0].get_metadata()[0].data.video_id
    for data in dataset:
        video_id = data.get_metadata()[0].data.video_id
        if _video_id != video_id:
            return True
    return False


class ActionOVDemoDataLoader(DataLoader):
    """DataLoader for online demo purpose.

    Since it is for online demo purpose it selects background frames from neighbor of key frame
    """

    def __init__(self, dataset: DatasetEntity, task_type: str, clip_len: int, width: int, height: int):
        self.task_type = task_type
        self.dataset = dataset
        self.clip_len = clip_len
        self.width = width
        self.height = height
        self.interval = 2

    def __len__(self):
        """Length of data loader."""
        return len(self.dataset)

    def __getitem__(self, index: int):
        """Sample frames from back and forth of key frame."""
        start = index - (self.clip_len // 2) * self.interval
        end = index + ((self.clip_len + 1) // 2) * self.interval
        frame_inds = list(range(start, end, self.interval))
        frame_inds = np.clip(frame_inds, 0, len(self.dataset) - 1)
        dataset_items = []
        for idx in frame_inds:
            dataset_item = self.dataset[int(idx)]
            dataset_items.append(dataset_item)
        return dataset_items

    def add_prediction(self, data: List[DatasetItemEntity], prediction: AnnotationSceneEntity):
        """Add prediction results to key frame.

        From sampling methods, we know that data[len(data) // 2] is key frame
        """
        dataset_item = data[len(data) // 2]
        if self.task_type == "ACTION_CLASSIFICATION":
            dataset_item.append_labels(prediction.annotations[0].get_labels())
        else:
            dataset_item.append_annotations(prediction.annotations)


class ActionOVClsDataLoader(DataLoader):
    """DataLoader for evaluation of action classification models.

    It iterates through clustered video, and it samples frames from given video
    """

    def __init__(self, dataset: DatasetEntity, clip_len: int, width: int, height: int):
        self.clip_len = clip_len
        self.width = width
        self.height = height

        video_info: Dict[str, List[DatasetItemEntity]] = {}
        for dataset_item in dataset:
            metadata = dataset_item.get_metadata()[0].data
            video_id = metadata.video_id
            if video_id in video_info:
                video_info[video_id].append(dataset_item)
            else:
                video_info[video_id] = [dataset_item]
        self.dataset = list(video_info.values())

        self.interval = 4

    def __len__(self):
        """Length of data loader."""
        return len(self.dataset)

    def __getitem__(self, index: int):
        """Sample frames from given video."""
        items = self.dataset[index]
        indices = self._get_indices(len(items))
        dataset_items = []
        for idx in indices:
            dataset_item = items[idx]
            dataset_items.append(dataset_item)
        return dataset_items

    def _get_indices(self, video_len: int):
        """Sample frame indices from video length."""
        ori_clip_len = self.clip_len * self.interval
        if video_len > ori_clip_len - 1:
            start = (video_len - ori_clip_len + 1) / 2
        else:
            start = 0
        frame_inds = np.arange(self.clip_len) * self.interval + int(start)
        frame_inds = np.mod(frame_inds, video_len)
        frame_inds = frame_inds.astype(np.int)
        return frame_inds

    def add_prediction(self, dataset: DatasetEntity, data: List[DatasetItemEntity], prediction: AnnotationSceneEntity):
        """Add prediction to dataset.

        Add prediction result to dataset_item in dataset, which has same video id with video data.
        """
        video_id = data[0].get_metadata()[0].data.video_id
        for dataset_item in dataset:
            if dataset_item.get_metadata()[0].data.video_id == video_id:
                dataset_item.append_labels(prediction.annotations[0].get_labels())


class ActionOVDetDataLoader(DataLoader):
    """DataLoader for evaluation of spatio-temporal action detection models.

    It iterates through DatasetEntity, which only contains non-empty frame(frame with actor annotation)
    It samples background frames from original DatasetEntity, which contain both empty frame and non-empty frame
    """

    def __init__(self, dataset: DatasetEntity, clip_len: int, width: int, height: int):
        self.original_dataset = dataset
        self.clip_len = clip_len
        self.width = width
        self.height = height

        self.dataset = deepcopy(dataset)
        video_info: Dict[str, Dict[str, int]] = {}
        for idx, dataset_item in enumerate(self.dataset):
            metadata = dataset_item.get_metadata()[0].data
            video_id = metadata.video_id
            timestamp = metadata.frame_idx
            if video_id in video_info:
                if video_info[video_id]["timestamp_start"] > timestamp:
                    video_info[video_id]["timestamp_start"] = timestamp
                if video_info[video_id]["timestamp_end"] < timestamp:
                    video_info[video_id]["timestamp_end"] = timestamp
            else:
                video_info[video_id] = {
                    "start_index": idx,
                    "timestamp_start": timestamp,
                    "timestamp_end": timestamp,
                }
        remove_indices = []
        for idx, dataset_item in enumerate(self.dataset):
            metadata = dataset_item.get_metadata()[0].data
            if metadata.is_empty_frame:
                remove_indices.append(idx)
                continue
            metadata.update("start_index", video_info[metadata.video_id]["start_index"])
            metadata.update("timestamp_start", video_info[metadata.video_id]["timestamp_start"])
            metadata.update("timestamp_end", video_info[metadata.video_id]["timestamp_end"])
        self.dataset.remove_at_indices(remove_indices)

        self.interval = 2
        self.fps = 1

    def __len__(self):
        """Length of data loader."""
        return len(self.dataset)

    def __getitem__(self, index: int):
        """Sample frames from back and forth of key frame, and all frames are from same video with key frame."""
        metadata = self.dataset[index].get_metadata()[0].data
        timestamp = metadata.frame_idx
        timestamp_start = metadata.timestamp_start
        timestamp_end = metadata.timestamp_end
        indices = self._get_indices(timestamp, timestamp_start, timestamp_end)
        indices = indices - timestamp_start + metadata.start_index
        dataset_items = []
        for idx in indices:
            dataset_item = self.original_dataset[int(idx)]
            dataset_items.append(dataset_item)
        return dataset_items

    def _get_indices(self, timestamp: int, timestamp_start: int, timestamp_end: int):
        """Get indices from timestamp.

        Samples from back and forth of key timestamp, and clips using start, and end timestamp
        """
        start = timestamp - (self.clip_len // 2) * self.interval
        end = timestamp + ((self.clip_len + 1) // 2) * self.interval
        frame_inds = list(range(start, end, self.interval))
        frame_inds = np.clip(frame_inds, timestamp_start, timestamp_end)
        return frame_inds

    def add_prediction(self, data: List[DatasetItemEntity], prediction: AnnotationSceneEntity):
        """Add prediction results to key frame."""
        dataset_item = data[len(data) // 2]
        dataset_item.append_annotations(prediction.annotations)



"""EfficientNet-V2 for hierarchical config."""

# pylint: disable=invalid-name

_base_ = "../base/models/efficientnet_v2.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(version="s_21k"),
    head=dict(
        type="CustomHierarchicalLinearClsHead",
        multilabel_loss=dict(
            type="AsymmetricLossWithIgnore",
            gamma_pos=0.0,
            gamma_neg=4.0,
        ),
    ),
)



"""This module contains the interface class for tasks that can deploy their models."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc

from otx.api.entities.model import ModelEntity


class IDeploymentTask(metaclass=abc.ABCMeta):
    """A base interface class for tasks which can deploy their models."""

    @abc.abstractmethod
    def deploy(self, output_model: ModelEntity) -> None:
        """This method defines the interface for deploy.

        Args:
            output_model: Output model
        """
        raise NotImplementedError



_base_ = ["./train.py", "../_base_/data/coco_ubt.py", "../_base_/models/detectors/detector.py"]

task_adapt = dict(
    type="mpa",
    op="REPLACE",
    efficient_mode=False,
)

custom_hooks = [
    dict(
        type="UnbiasedTeacherHook",
        epoch_momentum=0.1,
        start_epoch=2,
        # min_pseudo_label_ratio=0.1,
        min_pseudo_label_ratio=0.0,
    ),
    dict(
        type="DualModelEMAHook",
        epoch_momentum=0.4,
        start_epoch=2,
    ),
    dict(
        type="LazyEarlyStoppingHook",
        start=3,
        patience=5,
        iteration_patience=1000,
        metric="bbox_mAP",
        interval=1,
        priority=75,
    ),
]



"""This module contains the ModelLifecycle Enum."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from enum import Enum, auto


class ModelLifecycle(Enum):
    """This Enum represents the different stages in the ModelLifecycle.

    It is used by configuration parameters to indicate
    in which stage of the model lifecycle the parameter takes effect. Selecting a stage early in the lifecycle implies
    that all downstream stages are affected as well (e.g. if this is set to `ModelLifecycle.TRAINING`, it is assumed
    that inference and testing are also impacted).

    Currently the following stages are possible:
    ARCHITECTURE - Select this stage if the parameter modifies the model architecture, such that the most recently
        trained weights cannot directly by used for the next training round due to a model topology mismatch. For
        example, a parameter `model_depth` that controls the number of downsampling steps in a UNet model should
        have this stage set.
    TRAINING - Select this stage if the parameter is likely to change the outcome of the training process. For example,
        the parameter `learning_rate` should have this stage set.
    INFERENCE - Select this stage if the parameter changes the result of inference. For example, a parameter
        `probability_threshold` that controls the threshold for binary classification should have this stage set.
    TESTING - Select this stage if the parameter changes the outcome of the evaluation process. For example, a parameter
        'test_metric` that controls which metric to use for testing does not change training or inference results, but
        does affect the final evaluation of the model. Therefore, it should have this stage set.
    NONE - Select this stage if the parameter is non-functional, for example if it only impacts training speed but
        should not change the training outcome. For example, a parameter `num_workers` that controls the number of
        threads used in a dataloader should have this stage set.
    """

    NONE = auto()
    ARCHITECTURE = auto()
    TRAINING = auto()
    INFERENCE = auto()
    TESTING = auto()

    def __str__(self):
        """Retrieves the string representation of an instance of the Enum."""
        return self.name



"""This module contains Enums used in the configurable parameters within the OTX."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .auto_hpo_state import AutoHPOState
from .config_element_type import ConfigElementType
from .model_lifecycle import ModelLifecycle

__all__ = ["ConfigElementType", "ModelLifecycle", "AutoHPOState"]



"""OTX CLI entry point."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import argparse
import sys

from otx.cli.utils import telemetry

from .build import main as otx_build
from .demo import main as otx_demo
from .deploy import main as otx_deploy
from .eval import main as otx_eval
from .explain import main as otx_explain
from .export import main as otx_export
from .find import main as otx_find
from .optimize import main as otx_optimize
from .train import main as otx_train

__all__ = [
    "otx_demo",
    "otx_deploy",
    "otx_eval",
    "otx_explain",
    "otx_export",
    "otx_find",
    "otx_train",
    "otx_optimize",
    "otx_build",
]


def parse_args():
    """Parses command line arguments."""

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("operation", choices=[x[4:] for x in __all__], type=str)

    return parser.parse_known_args()[0]


def main():
    """Entry point for OTX CLI.

    This function is a single entry point for all OTX CLI related operations:
      - build
      - demo
      - deploy
      - eval
      - explain
      - export
      - find
      - train
      - optimize
    """

    name = parse_args().operation
    sys.argv[0] = f"otx {name}"
    del sys.argv[1]

    tm_session = telemetry.init_telemetry_session()
    results = {}
    try:
        results = globals()[f"otx_{name}"]()
        if results is None:
            results = dict(retcode=0)
    except Exception as error:
        results["retcode"] = -1
        results["exception"] = repr(error)
        telemetry.send_cmd_results(tm_session, name, results)
        raise
    else:
        telemetry.send_cmd_results(tm_session, name, results)
    finally:
        telemetry.close_telemetry_session(tm_session)

    return results.get("retcode", 0)


if __name__ == "__main__":
    main()



"""Implementation of EfficientNet.

Original papers:
- 'EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks,' https://arxiv.org/abs/1905.11946,
- 'Adversarial Examples Improve Image Recognition,' https://arxiv.org/abs/1911.09665.
"""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
# pylint: disable=too-many-arguments, invalid-name, unused-argument, too-many-lines
# pylint: disable=too-many-instance-attributes,too-many-statements, too-many-branches, too-many-locals

import math
import os

import torch.nn.functional as F
from mmcv.cnn import build_activation_layer
from mmcv.cnn.bricks import ConvModule
from mmcv.runner import load_checkpoint
from torch import nn
from torch.nn import init

from otx.mpa.utils.logger import get_logger

from ..builder import BACKBONES

logger = get_logger()

PRETRAINED_ROOT = "https://github.com/osmr/imgclsmob/releases/download/v0.0.364/"
pretrained_urls = {
    "efficientnet_b0": PRETRAINED_ROOT + "efficientnet_b0-0752-0e386130.pth.zip",
}


def conv1x1_block(
    in_channels,
    out_channels,
    stride=1,
    padding=0,
    groups=1,
    bias=False,
    use_bn=True,
    bn_eps=1e-5,
    activation="ReLU",
):
    """Conv block."""

    return ConvModule(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=1,
        stride=stride,
        padding=padding,
        groups=groups,
        bias=bias,
        norm_cfg=(dict(type="BN", eps=bn_eps) if use_bn else None),
        act_cfg=(dict(type=activation) if activation else None),
    )


def conv3x3_block(
    in_channels,
    out_channels,
    stride=1,
    padding=1,
    dilation=1,
    groups=1,
    bias=False,
    use_bn=True,
    bn_eps=1e-5,
    activation="ReLU",
    IN_conv=False,
):
    """Conv block."""

    return ConvModule(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        bias=bias,
        norm_cfg=(dict(type="BN", eps=bn_eps) if use_bn else None),
        act_cfg=(dict(type=activation) if activation else None),
    )


def dwconv3x3_block(
    in_channels,
    out_channels,
    stride=1,
    padding=1,
    dilation=1,
    bias=False,
    use_bn=True,
    bn_eps=1e-5,
    activation="ReLU",
):
    """Conv block."""

    return ConvModule(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=3,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=out_channels,
        bias=bias,
        norm_cfg=(dict(type="BN", eps=bn_eps) if use_bn else None),
        act_cfg=(dict(type=activation) if activation else None),
    )


def dwconv5x5_block(
    in_channels,
    out_channels,
    stride=1,
    padding=2,
    dilation=1,
    bias=False,
    use_bn=True,
    bn_eps=1e-5,
    activation="ReLU",
):
    """Conv block."""
    return ConvModule(
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_size=5,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=out_channels,
        bias=bias,
        norm_cfg=(dict(type="BN", eps=bn_eps) if use_bn else None),
        act_cfg=(dict(type=activation) if activation else None),
    )


def round_channels(channels, divisor=8):
    """Round weighted channel number (make divisible operation).

    Args:
        channels : int or float. Original number of channels.
        divisor : int, default 8. Alignment value.
    """
    rounded_channels = max(int(channels + divisor / 2.0) // divisor * divisor, divisor)
    if float(rounded_channels) < 0.9 * channels:
        rounded_channels += divisor
    return rounded_channels


def calc_tf_padding(x, kernel_size, stride=1, dilation=1):
    """Calculate TF-same like padding size.

    Args:
        x : tensor. Input tensor.
        kernel_size : int. Convolution window size.
        stride : int, default 1. Strides of the convolution.
        dilation : int, default 1. Dilation value for convolution layer.
    """
    height, width = x.size()[2:]
    oh = math.ceil(height / stride)
    ow = math.ceil(width / stride)
    pad_h = max((oh - 1) * stride + (kernel_size - 1) * dilation + 1 - height, 0)
    pad_w = max((ow - 1) * stride + (kernel_size - 1) * dilation + 1 - width, 0)
    return pad_h // 2, pad_h - pad_h // 2, pad_w // 2, pad_w - pad_w // 2


class SEBlock(nn.Module):
    """Squeeze-and-Excitation block from 'Squeeze-and-Excitation Networks,'.

    https://arxiv.org/abs/1709.01507.

    Args:
        channels : int. Number of channels.
        reduction : int, default 16. Squeeze reduction value.
        mid_channels : int or None, default None. Number of middle channels.
        round_mid : bool, default False. Whether to round middle channel number (make divisible by 8).
        use_conv : bool, default True. Whether to convolutional layers instead of fully-connected ones.
        activation : function, or str, or nn.Module, default 'relu'. Activation function after the first convolution.
        out_activation : function, or str, or nn.Module, Activation function after the last convolution.
    """

    def __init__(
        self,
        channels,
        reduction=16,
        mid_channels=None,
        round_mid=False,
        use_conv=True,
        mid_activation="ReLU",
        out_activation="Sigmoid",
    ):
        super().__init__()
        self.use_conv = use_conv
        if mid_channels is None:
            mid_channels = channels // reduction if not round_mid else round_channels(float(channels) / reduction)

        self.pool = nn.AdaptiveAvgPool2d(output_size=1)
        if use_conv:
            self.conv1 = nn.Conv2d(
                in_channels=channels,
                out_channels=mid_channels,
                kernel_size=1,
                stride=1,
                groups=1,
                bias=True,
            )
        else:
            self.fc1 = nn.Linear(in_features=channels, out_features=mid_channels)
        self.activ = build_activation_layer(dict(type=mid_activation))
        if use_conv:
            self.conv2 = nn.Conv2d(
                in_channels=mid_channels,
                out_channels=channels,
                kernel_size=1,
                stride=1,
                groups=1,
                bias=True,
            )
        else:
            self.fc2 = nn.Linear(in_features=mid_channels, out_features=channels)
        self.sigmoid = build_activation_layer(dict(type=out_activation))

    def forward(self, x):
        """Forward."""
        w = self.pool(x)
        if not self.use_conv:
            w = w.view(x.size(0), -1)
        w = self.conv1(w) if self.use_conv else self.fc1(w)
        w = self.activ(w)
        w = self.conv2(w) if self.use_conv else self.fc2(w)
        w = self.sigmoid(w)
        if not self.use_conv:
            w = w.unsqueeze(2).unsqueeze(3)
        x = x * w
        return x


class EffiDwsConvUnit(nn.Module):
    """EfficientNet specific depthwise separable conv block/unit with BatchNorms and activations at each conv.

    Args:
        in_channels : int. Number of input channels.
        out_channels : int. Number of output channels.
        stride : int or tuple/list of 2 int. Strides of the second convolution layer.
        bn_eps : float. Small float added to variance in Batch norm.
        activation : str. Name of activation function.
        tf_mode : bool. Whether to use TF-like mode.
    """

    def __init__(self, in_channels, out_channels, stride, bn_eps, activation, tf_mode):
        super().__init__()
        self.tf_mode = tf_mode
        self.residual = (in_channels == out_channels) and (stride == 1)

        self.dw_conv = dwconv3x3_block(
            in_channels=in_channels,
            out_channels=in_channels,
            padding=(0 if tf_mode else 1),
            bn_eps=bn_eps,
            activation=activation,
        )
        self.se = SEBlock(channels=in_channels, reduction=4, mid_activation=activation)
        self.pw_conv = conv1x1_block(
            in_channels=in_channels,
            out_channels=out_channels,
            bn_eps=bn_eps,
            activation=None,
        )

    def forward(self, x):
        """Forward."""
        if self.residual:
            identity = x
        if self.tf_mode:
            x = F.pad(x, pad=calc_tf_padding(x, kernel_size=3))
        x = self.dw_conv(x)
        x = self.se(x)
        x = self.pw_conv(x)
        if self.residual:
            x = x + identity
        return x


class EffiInvResUnit(nn.Module):
    """EfficientNet inverted residual unit.

    Args:
        in_channels : int. Number of input channels.
        out_channels : int. Number of output channels.
        kernel_size : int or tuple/list of 2 int. Convolution window size.
        stride : int or tuple/list of 2 int. Strides of the second convolution layer.
        exp_factor : int. Factor for expansion of channels.
        se_factor : int. SE reduction factor for each unit.
        bn_eps : float. Small float added to variance in Batch norm.
        activation : str. Name of activation function.
        tf_mode : bool. Whether to use TF-like mode.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride,
        exp_factor,
        se_factor,
        bn_eps,
        activation,
        tf_mode,
    ):
        super().__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.tf_mode = tf_mode
        self.residual = (in_channels == out_channels) and (stride == 1)
        self.use_se = se_factor > 0
        mid_channels = in_channels * exp_factor
        dwconv_block_fn = dwconv3x3_block if kernel_size == 3 else (dwconv5x5_block if kernel_size == 5 else None)

        self.conv1 = conv1x1_block(
            in_channels=in_channels,
            out_channels=mid_channels,
            bn_eps=bn_eps,
            activation=activation,
        )
        self.conv2 = dwconv_block_fn(
            in_channels=mid_channels,
            out_channels=mid_channels,
            stride=stride,
            padding=(0 if tf_mode else kernel_size // 2),
            bn_eps=bn_eps,
            activation=activation,
        )
        if self.use_se:
            self.se = SEBlock(
                channels=mid_channels,
                reduction=(exp_factor * se_factor),
                mid_activation=activation,
            )
        self.conv3 = conv1x1_block(
            in_channels=mid_channels,
            out_channels=out_channels,
            bn_eps=bn_eps,
            activation=None,
        )

    def forward(self, x):
        """Forward."""
        if self.residual:
            identity = x
        x = self.conv1(x)
        if self.tf_mode:
            x = F.pad(
                x,
                pad=calc_tf_padding(x, kernel_size=self.kernel_size, stride=self.stride),
            )
        x = self.conv2(x)
        if self.use_se:
            x = self.se(x)
        x = self.conv3(x)
        if self.residual:
            x = x + identity
        return x


class EffiInitBlock(nn.Module):
    """EfficientNet specific initial block.

    Args:
        in_channels : int. Number of input channels.
        out_channels : int. Number of output channels.
        bn_eps : float. Small float added to variance in Batch norm.
        activation : str. Name of activation function.
        tf_mode : bool. Whether to use TF-like mode.
    """

    def __init__(self, in_channels, out_channels, bn_eps, activation, tf_mode, IN_conv1):
        super().__init__()
        self.tf_mode = tf_mode

        self.conv = conv3x3_block(
            in_channels=in_channels,
            out_channels=out_channels,
            stride=2,
            padding=(0 if tf_mode else 1),
            bn_eps=bn_eps,
            activation=activation,
            IN_conv=IN_conv1,
        )

    def forward(self, x):
        """Forward."""
        if self.tf_mode:
            x = F.pad(x, pad=calc_tf_padding(x, kernel_size=3, stride=2))
        x = self.conv(x)
        return x


class EfficientNet(nn.Module):
    """EfficientNet.

    Args:
        channels : list of list of int. Number of output channels for each unit.
        init_block_channels : int. Number of output channels for initial unit.
        final_block_channels : int. Number of output channels for the final block of the feature extractor.
        kernel_sizes : list of list of int. Number of kernel sizes for each unit.
        strides_per_stage : list int. Stride value for the first unit of each stage.
        expansion_factors : list of list of int. Number of expansion factors for each unit.
        dropout_rate : float, default 0.2. Fraction of the input units to drop. Must be a number between 0 and 1.
        tf_mode : bool, default False. Whether to use TF-like mode.
        bn_eps : float, default 1e-5. Small float added to variance in Batch norm.
        in_channels : int, default 3. Number of input channels.
        in_size : tuple of two ints, default (224, 224). Spatial size of the expected input image.
        num_classes : int, default 1000. Number of classification classes.
    """

    def __init__(
        self,
        channels,
        init_block_channels,
        final_block_channels,
        kernel_sizes,
        strides_per_stage,
        expansion_factors,
        tf_mode=False,
        bn_eps=1e-5,
        in_channels=3,
        in_size=(224, 224),
        dropout_cls=None,
        pooling_type="avg",
        bn_eval=False,
        bn_frozen=False,
        IN_first=False,
        IN_conv1=False,
        pretrained=False,
        **kwargs,
    ):

        super().__init__(**kwargs)
        self.num_classes = 1000
        self.pretrained = pretrained
        self.in_size = in_size
        self.input_IN = nn.InstanceNorm2d(3, affine=True) if IN_first else None
        self.bn_eval = bn_eval
        self.bn_frozen = bn_frozen
        self.pooling_type = pooling_type
        self.num_features = self.num_head_features = final_block_channels
        activation = "Swish"
        self.features = nn.Sequential()
        self.features.add_module(
            "init_block",
            EffiInitBlock(
                in_channels=in_channels,
                out_channels=init_block_channels,
                bn_eps=bn_eps,
                activation=activation,
                tf_mode=tf_mode,
                IN_conv1=IN_conv1,
            ),
        )
        in_channels = init_block_channels
        for i, channels_per_stage in enumerate(channels):
            kernel_sizes_per_stage = kernel_sizes[i]
            expansion_factors_per_stage = expansion_factors[i]
            stage = nn.Sequential()
            for j, out_channels in enumerate(channels_per_stage):
                kernel_size = kernel_sizes_per_stage[j]
                expansion_factor = expansion_factors_per_stage[j]
                stride = strides_per_stage[i] if (j == 0) else 1
                if i == 0:
                    stage.add_module(
                        f"unit{j + 1}",
                        EffiDwsConvUnit(
                            in_channels=in_channels,
                            out_channels=out_channels,
                            stride=stride,
                            bn_eps=bn_eps,
                            activation=activation,
                            tf_mode=tf_mode,
                        ),
                    )
                else:
                    stage.add_module(
                        f"unit{j + 1}",
                        EffiInvResUnit(
                            in_channels=in_channels,
                            out_channels=out_channels,
                            kernel_size=kernel_size,
                            stride=stride,
                            exp_factor=expansion_factor,
                            se_factor=4,
                            bn_eps=bn_eps,
                            activation=activation,
                            tf_mode=tf_mode,
                        ),
                    )
                in_channels = out_channels
            self.features.add_module(f"stage{i+1}", stage)
            # activation = activation if self.loss == 'softmax': else lambda: nn.PReLU(init=0.25)
        self.features.add_module(
            "final_block",
            conv1x1_block(
                in_channels=in_channels, out_channels=final_block_channels, bn_eps=bn_eps, activation=activation
            ),
        )
        self._init_params()

    def _init_params(self):
        for module in self.named_modules():
            if isinstance(module, nn.Conv2d):
                init.kaiming_uniform_(module.weight)
                if module.bias is not None:
                    init.constant_(module.bias, 0)

    def forward(self, x, return_featuremaps=False, get_embeddings=False):
        """Forward."""
        if self.input_IN is not None:
            x = self.input_IN(x)  # pylint: disable=not-callable

        y = self.features(x)
        if return_featuremaps:
            return y

        glob_features = self._glob_feature_vector(y, self.pooling_type, reduce_dims=False)

        logits = self.output(glob_features.view(x.shape[0], -1))

        if not self.training and self.classification:
            return [logits]

        if get_embeddings:
            out_data = [logits, glob_features.view(x.shape[0], -1)]
        elif self.loss in ["softmax", "am_softmax"]:
            if self.lr_finder.enable and self.lr_finder.mode == "automatic":
                out_data = logits
            else:
                out_data = [logits]

        elif self.loss in ["triplet"]:
            out_data = [logits, glob_features]
        else:
            raise KeyError(f"Unsupported loss: {self.loss}")

        if self.lr_finder.enable and self.lr_finder.mode == "automatic":
            return out_data
        return tuple(out_data)


def get_efficientnet(
    version,
    in_size,
    tf_mode=False,
    bn_eps=1e-5,
    model_name=None,
    pretrained=False,
    root=os.path.join("~", ".torch", "models"),
    **kwargs,
):
    """Create EfficientNet model with specific parameters.

    Args:
        version : str. Version of EfficientNet ('b0'...'b8').
        in_size : tuple of two ints. Spatial size of the expected input image.
        tf_mode : bool, default False. Whether to use TF-like mode.
        bn_eps : float, default 1e-5. Small float added to variance in Batch norm.
        model_name : str or None, default None. Model name for loading pretrained model.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    if version == "b0":
        assert in_size == (224, 224)
        depth_factor = 1.0
        width_factor = 1.0
        dropout_rate = 0.2
    elif version == "b1":
        assert in_size == (240, 240)
        depth_factor = 1.1
        width_factor = 1.0
        dropout_rate = 0.2
    elif version == "b2":
        assert in_size == (260, 260)
        depth_factor = 1.2
        width_factor = 1.1
        dropout_rate = 0.3
    elif version == "b3":
        assert in_size == (300, 300)
        depth_factor = 1.4
        width_factor = 1.2
        dropout_rate = 0.3
    elif version == "b4":
        assert in_size == (380, 380)
        depth_factor = 1.8
        width_factor = 1.4
        dropout_rate = 0.4
    elif version == "b5":
        assert in_size == (456, 456)
        depth_factor = 2.2
        width_factor = 1.6
        dropout_rate = 0.4
    elif version == "b6":
        assert in_size == (528, 528)
        depth_factor = 2.6
        width_factor = 1.8
        dropout_rate = 0.5
    elif version == "b7":
        assert in_size == (600, 600)
        depth_factor = 3.1
        width_factor = 2.0
        dropout_rate = 0.5
    elif version == "b8":
        assert in_size == (672, 672)
        depth_factor = 3.6
        width_factor = 2.2
        dropout_rate = 0.5
    else:
        raise ValueError(f"Unsupported EfficientNet version {version}")

    init_block_channels = 32
    layers = [1, 2, 2, 3, 3, 4, 1]
    downsample = [1, 1, 1, 1, 0, 1, 0]
    channels_per_layers = [16, 24, 40, 80, 112, 192, 320]
    expansion_factors_per_layers = [1, 6, 6, 6, 6, 6, 6]
    kernel_sizes_per_layers = [3, 3, 5, 3, 5, 5, 3]
    strides_per_stage = [1, 2, 2, 2, 1, 2, 1]
    final_block_channels = 1280

    layers = [int(math.ceil(li * depth_factor)) for li in layers]
    channels_per_layers = [round_channels(ci * width_factor) for ci in channels_per_layers]

    from functools import reduce

    channels = reduce(
        lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
        zip(channels_per_layers, layers, downsample),
        [],
    )
    kernel_sizes = reduce(
        lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
        zip(kernel_sizes_per_layers, layers, downsample),
        [],
    )
    expansion_factors = reduce(
        lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
        zip(expansion_factors_per_layers, layers, downsample),
        [],
    )
    strides_per_stage = reduce(
        lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
        zip(strides_per_stage, layers, downsample),
        [],
    )
    strides_per_stage = [si[0] for si in strides_per_stage]

    init_block_channels = round_channels(init_block_channels * width_factor)

    if width_factor > 1.0:
        assert int(final_block_channels * width_factor) == round_channels(final_block_channels * width_factor)
        final_block_channels = round_channels(final_block_channels * width_factor)

    net = EfficientNet(
        channels=channels,
        init_block_channels=init_block_channels,
        final_block_channels=final_block_channels,
        kernel_sizes=kernel_sizes,
        strides_per_stage=strides_per_stage,
        expansion_factors=expansion_factors,
        dropout_rate=dropout_rate,
        tf_mode=tf_mode,
        bn_eps=bn_eps,
        in_size=in_size,
        **kwargs,
    )

    return net


def efficientnet_b0(in_size=(224, 224), **kwargs):
    """EfficientNet-B0.

    Args:
        in_size : tuple of two ints, default (224, 224). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b0", in_size=in_size, model_name="efficientnet_b0", **kwargs)


def efficientnet_b1(in_size=(240, 240), **kwargs):
    """EfficientNet-B1.

    Args:
        in_size : tuple of two ints, default (240, 240). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b1", in_size=in_size, model_name="efficientnet_b1", **kwargs)


def efficientnet_b2(in_size=(260, 260), **kwargs):
    """EfficientNet-B2.

    Args:
        in_size : tuple of two ints, default (260, 260). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b2", in_size=in_size, model_name="efficientnet_b2", **kwargs)


def efficientnet_b3(in_size=(300, 300), **kwargs):
    """EfficientNet-B3.

    Args:
        in_size : tuple of two ints, default (300, 300). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b3", in_size=in_size, model_name="efficientnet_b3", **kwargs)


def efficientnet_b4(in_size=(380, 380), **kwargs):
    """EfficientNet-B4.

    Args:
        in_size : tuple of two ints, default (380, 380). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b4", in_size=in_size, model_name="efficientnet_b4", **kwargs)


def efficientnet_b5(in_size=(456, 456), **kwargs):
    """EfficientNet-B5.

    Args:
        in_size : tuple of two ints, default (456, 456). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b5", in_size=in_size, model_name="efficientnet_b5", **kwargs)


def efficientnet_b6(in_size=(528, 528), **kwargs):
    """EfficientNet-B6 model.

    Args:
        in_size : tuple of two ints, default (528, 528). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b6", in_size=in_size, model_name="efficientnet_b6", **kwargs)


def efficientnet_b7(in_size=(600, 600), **kwargs):
    """EfficientNet-B7 model.

    Args:
        in_size : tuple of two ints, default (600, 600). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b7", in_size=in_size, model_name="efficientnet_b7", **kwargs)


def efficientnet_b8(in_size=(672, 672), **kwargs):
    """EfficientNet-B8.

    Args:
        in_size : tuple of two ints, default (672, 672). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(version="b8", in_size=in_size, model_name="efficientnet_b8", **kwargs)


def efficientnet_b0b(in_size=(224, 224), **kwargs):
    """EfficientNet-B0-b.

    Args:
        in_size : tuple of two ints, default (224, 224). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b0",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b0b",
        **kwargs,
    )


def efficientnet_b1b(in_size=(240, 240), **kwargs):
    """EfficientNet-B1-b.

    Args:
        in_size : tuple of two ints, default (240, 240). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b1",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b1b",
        **kwargs,
    )


def efficientnet_b2b(in_size=(260, 260), **kwargs):
    """EfficientNet-B2-b.

    Args:
        in_size : tuple of two ints, default (260, 260). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b2",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b2b",
        **kwargs,
    )


def efficientnet_b3b(in_size=(300, 300), **kwargs):
    """EfficientNet-B3-b.

    Args:
        in_size : tuple of two ints, default (300, 300). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b3",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b3b",
        **kwargs,
    )


def efficientnet_b4b(in_size=(380, 380), **kwargs):
    """EfficientNet-B4-b.

    Args:
        in_size : tuple of two ints, default (380, 380). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b4",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b4b",
        **kwargs,
    )


def efficientnet_b5b(in_size=(456, 456), **kwargs):
    """EfficientNet-B5-b.

    Args:
        in_size : tuple of two ints, default (456, 456). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b5",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b5b",
        **kwargs,
    )


def efficientnet_b6b(in_size=(528, 528), **kwargs):
    """EfficientNet-B6-b.

    Args:
        in_size : tuple of two ints, default (528, 528). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b6",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b6b",
        **kwargs,
    )


def efficientnet_b7b(in_size=(600, 600), **kwargs):
    """EfficientNet-B7-b.

    Args:
        in_size : tuple of two ints, default (600, 600). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b7",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b7b",
        **kwargs,
    )


def efficientnet_b0c(in_size=(224, 224), **kwargs):
    """EfficientNet-B0-c.

    Args:
        in_size : tuple of two ints, default (224, 224). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b0",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b0c",
        **kwargs,
    )


def efficientnet_b1c(in_size=(240, 240), **kwargs):
    """EfficientNet-B1-c.

    Args:
        in_size : tuple of two ints, default (240, 240). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b1",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b1c",
        **kwargs,
    )


def efficientnet_b2c(in_size=(260, 260), **kwargs):
    """EfficientNet-B2-c.

    Args:
        in_size : tuple of two ints, default (260, 260). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b2",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b2c",
        **kwargs,
    )


def efficientnet_b3c(in_size=(300, 300), **kwargs):
    """EfficientNet-B3-c.

    Args:
        in_size : tuple of two ints, default (300, 300). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b3",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b3c",
        **kwargs,
    )


def efficientnet_b4c(in_size=(380, 380), **kwargs):
    """EfficientNet-B4-c.

    Args:
        in_size : tuple of two ints, default (380, 380). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b4",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b4c",
        **kwargs,
    )


def efficientnet_b5c(in_size=(456, 456), **kwargs):
    """EfficientNet-B5-c.

    Args:
        in_size : tuple of two ints, default (456, 456). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b5",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b5c",
        **kwargs,
    )


def efficientnet_b6c(in_size=(528, 528), **kwargs):
    """EfficientNet-B6-c.

    Args:
        in_size : tuple of two ints, default (528, 528). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b6",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b6c",
        **kwargs,
    )


def efficientnet_b7c(in_size=(600, 600), **kwargs):
    """EfficientNet-B7-c.

    Args:
        in_size : tuple of two ints, default (600, 600). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """
    return get_efficientnet(
        version="b7",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b7c",
        **kwargs,
    )


def efficientnet_b8c(in_size=(672, 672), **kwargs):
    """EfficientNet-B8-c.

    Args:
        in_size : tuple of two ints, default (672, 672). Spatial size of the expected input image.
        pretrained : bool, default False. Whether to load the pretrained weights for model.
        root : str, default '~/.torch/models'. Location for keeping the model parameters.
    """

    return get_efficientnet(
        version="b8",
        in_size=in_size,
        tf_mode=True,
        bn_eps=1e-3,
        model_name="efficientnet_b8c",
        **kwargs,
    )


def _calc_width(net):
    import numpy as np

    net_params = filter(lambda p: p.requires_grad, net.parameters())
    weight_count = 0
    for param in net_params:
        weight_count += np.prod(param.size())
    return weight_count


def _test():
    import torch

    pretrained = False

    models = [
        efficientnet_b0,
        efficientnet_b1,
        efficientnet_b2,
        efficientnet_b3,
        efficientnet_b4,
        efficientnet_b5,
        efficientnet_b6,
        efficientnet_b7,
        efficientnet_b8,
        efficientnet_b0b,
        efficientnet_b1b,
        efficientnet_b2b,
        efficientnet_b3b,
        efficientnet_b4b,
        efficientnet_b5b,
        efficientnet_b6b,
        efficientnet_b7b,
        efficientnet_b0c,
        efficientnet_b1c,
        efficientnet_b2c,
        efficientnet_b3c,
        efficientnet_b4c,
        efficientnet_b5c,
        efficientnet_b6c,
        efficientnet_b7c,
        efficientnet_b8c,
    ]

    for model in models:

        net = model(pretrained=pretrained)

        # net.train()
        net.eval()
        weight_count = _calc_width(net)
        print(f"m={model.__name__}, {weight_count}")
        assert model != efficientnet_b0 or weight_count == 5288548
        assert model != efficientnet_b1 or weight_count == 7794184
        assert model != efficientnet_b2 or weight_count == 9109994
        assert model != efficientnet_b3 or weight_count == 12233232
        assert model != efficientnet_b4 or weight_count == 19341616
        assert model != efficientnet_b5 or weight_count == 30389784
        assert model != efficientnet_b6 or weight_count == 43040704
        assert model != efficientnet_b7 or weight_count == 66347960
        assert model != efficientnet_b8 or weight_count == 87413142
        assert model != efficientnet_b0b or weight_count == 5288548
        assert model != efficientnet_b1b or weight_count == 7794184
        assert model != efficientnet_b2b or weight_count == 9109994
        assert model != efficientnet_b3b or weight_count == 12233232
        assert model != efficientnet_b4b or weight_count == 19341616
        assert model != efficientnet_b5b or weight_count == 30389784
        assert model != efficientnet_b6b or weight_count == 43040704
        assert model != efficientnet_b7b or weight_count == 66347960

        x = torch.randn(1, 3, net.in_size[0], net.in_size[1])
        y = net(x)
        y.sum().backward()
        assert tuple(y.size()) == (1, 1000)


@BACKBONES.register_module()
class OTXEfficientNet(EfficientNet):
    """Create EfficientNet model with specific parameters.

    Args:
        version : str. Version of EfficientNet ('b0'...'b8').
        in_size : tuple of two ints. Spatial size of the expected input image.
    """

    def __init__(self, version, **kwargs):
        self.model_name = "efficientnet_" + version

        if version == "b0":
            in_size = (224, 224)
            depth_factor = 1.0
            width_factor = 1.0
        elif version == "b1":
            in_size = (240, 240)
            depth_factor = 1.1
            width_factor = 1.0
        elif version == "b2":
            in_size = (260, 260)
            depth_factor = 1.2
            width_factor = 1.1
        elif version == "b3":
            in_size = (300, 300)
            depth_factor = 1.4
            width_factor = 1.2
        elif version == "b4":
            in_size = (380, 380)
            depth_factor = 1.8
            width_factor = 1.4
        elif version == "b5":
            in_size = (456, 456)
            depth_factor = 2.2
            width_factor = 1.6
        elif version == "b6":
            in_size = (528, 528)
            depth_factor = 2.6
            width_factor = 1.8
        elif version == "b7":
            in_size = (600, 600)
            depth_factor = 3.1
            width_factor = 2.0
        elif version == "b8":
            in_size = (672, 672)
            depth_factor = 3.6
            width_factor = 2.2
        else:
            raise ValueError(f"Unsupported EfficientNet version {version}")

        init_block_channels = 32
        layers = [1, 2, 2, 3, 3, 4, 1]
        downsample = [1, 1, 1, 1, 0, 1, 0]
        channels_per_layers = [16, 24, 40, 80, 112, 192, 320]
        expansion_factors_per_layers = [1, 6, 6, 6, 6, 6, 6]
        kernel_sizes_per_layers = [3, 3, 5, 3, 5, 5, 3]
        strides_per_stage = [1, 2, 2, 2, 1, 2, 1]
        final_block_channels = 1280

        layers = [int(math.ceil(li * depth_factor)) for li in layers]
        channels_per_layers = [round_channels(ci * width_factor) for ci in channels_per_layers]

        from functools import reduce

        channels = reduce(
            lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
            zip(channels_per_layers, layers, downsample),
            [],
        )
        kernel_sizes = reduce(
            lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
            zip(kernel_sizes_per_layers, layers, downsample),
            [],
        )
        expansion_factors = reduce(
            lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
            zip(expansion_factors_per_layers, layers, downsample),
            [],
        )
        strides_per_stage = reduce(
            lambda x, y: x + [[y[0]] * y[1]] if y[2] != 0 else x[:-1] + [x[-1] + [y[0]] * y[1]],
            zip(strides_per_stage, layers, downsample),
            [],
        )
        strides_per_stage = [si[0] for si in strides_per_stage]

        init_block_channels = round_channels(init_block_channels * width_factor)

        if width_factor > 1.0:
            assert int(final_block_channels * width_factor) == round_channels(final_block_channels * width_factor)
            final_block_channels = round_channels(final_block_channels * width_factor)

        super().__init__(
            channels=channels,
            init_block_channels=init_block_channels,
            final_block_channels=final_block_channels,
            kernel_sizes=kernel_sizes,
            strides_per_stage=strides_per_stage,
            expansion_factors=expansion_factors,
            dropout_cls=dict(dist="none"),
            tf_mode=False,
            bn_eps=1e-5,
            in_size=in_size,
            **kwargs,
        )
        self.init_weights(self.pretrained)

    def forward(self, x):
        """Forward."""
        return super().forward(x, return_featuremaps=True)

    def init_weights(self, pretrained=None):
        """Initialize weights."""
        if isinstance(pretrained, str) and os.path.exists(pretrained):
            load_checkpoint(self, pretrained)
            logger.info(f"init weight - {pretrained}")
        elif pretrained is not None:
            load_checkpoint(self, pretrained_urls[self.model_name])
            logger.info(f"init weight - {pretrained_urls[self.model_name]}")



"""HPO algorithm abstract class."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import json
import logging
import tempfile
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, Dict, List, Optional, Union

from otx.algorithms.common.utils.utils import check_mode_input, check_positive
from otx.hpo.search_space import SearchSpace

logger = logging.getLogger(__name__)


class HpoBase(ABC):
    """Abstract class for HPO algorithms.

    This implements class for HPO algorithm base.
    Only common methods are implemented but not core algorithm of HPO.

    Args:
        search_space (Dict[str, Dict[str, Any]]): hyper parameter search space to find.
        save_path (Optional[str]): path where result of HPO is saved.
        mode (str, optinal): One of {min, max}. Determines whether objective is
                    minimizing or maximizing the metric attribute.
        num_trials (Optional[int]): How many training to conduct for HPO.
        num_workers (int): How many trains are executed in parallel.
        num_full_iterations (int): epoch for traninig after HPO.
        non_pure_train_ratio (float): ratio of validation time to (train time + validation time)
        full_dataset_size (int): train dataset size
        metric (str): Which score metric to use.
        expected_time_ratio (Optional[Union[int, float]]): Time to use for HPO.
                                            If HPO is configured automatically,
                                            HPO use time about exepected_time_ratio *
                                            train time after HPO times.
        maximum_resource (Optional[Union[int, float]]): Maximum resource to use for training each trial.
        subset_ratio (Optional[Union[float, int]]): ratio to how many train dataset to use for each trial.
                                     The lower value is, the faster the speed is.
                                     But If it's too low, HPO can be unstable.
        min_subset_size (int) : Minimum size of subset. Default value is 500.
        resume (bool): resume flag decide to use previous HPO results.
                       If HPO completed, you can just use optimized hyper parameters.
                       If HPO stopped in middle, you can resume in middle.
        prior_hyper_parameters (Optional[Union[Dict, List[Dict]]]) = Hyper parameters to try first.
        acceptable_additional_time_ratio (Union[float, int]) = Decide how much additional time can be acceptable.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        search_space: Dict[str, Dict[str, Any]],
        save_path: Optional[str] = None,
        mode: str = "max",
        num_trials: Optional[int] = None,
        num_workers: int = 1,
        num_full_iterations: Union[int, float] = 1,
        non_pure_train_ratio: float = 0.2,
        full_dataset_size: int = 0,
        metric: str = "mAP",
        expected_time_ratio: Optional[Union[int, float]] = None,
        maximum_resource: Optional[Union[int, float]] = None,
        subset_ratio: Optional[Union[float, int]] = None,
        min_subset_size: int = 500,
        resume: bool = False,
        prior_hyper_parameters: Optional[Union[Dict, List[Dict]]] = None,
        acceptable_additional_time_ratio: Union[float, int] = 1.0,
    ):
        # pylint: disable=too-many-arguments, too-many-locals
        check_mode_input(mode)
        check_positive(full_dataset_size, "full_dataset_size")
        check_positive(num_full_iterations, "num_full_iterations")
        if not 0 < non_pure_train_ratio <= 1:
            raise ValueError(
                "non_pure_train_ratio should be greater than 0 and lesser than or equal to 1."
                f" Your value is {subset_ratio}"
            )
        if maximum_resource is not None:
            check_positive(maximum_resource, "maximum_resource")
        if num_trials is not None:
            check_positive(num_trials, "num_trials")
        check_positive(num_workers, "num_workers")
        if subset_ratio is not None:
            if not 0 < subset_ratio <= 1:
                raise ValueError(
                    "subset_ratio should be greater than 0 and lesser than or equal to 1."
                    f" Your value is {subset_ratio}"
                )

        if save_path is None:
            save_path = tempfile.mkdtemp(prefix="OTX-hpo-")
        self.save_path = save_path
        self.search_space = SearchSpace(search_space)
        self.mode = mode
        self.num_trials = num_trials
        self.num_workers = num_workers
        self.num_full_iterations = num_full_iterations
        self.non_pure_train_ratio = non_pure_train_ratio
        self.full_dataset_size = full_dataset_size
        self.expected_time_ratio = expected_time_ratio
        self.maximum_resource = maximum_resource
        self.subset_ratio = subset_ratio
        self.min_subset_size = min_subset_size
        self.resume = resume
        self.hpo_status: dict = {}
        self.metric = metric
        self.acceptable_additional_time_ratio = acceptable_additional_time_ratio
        if prior_hyper_parameters is None:
            prior_hyper_parameters = []
        elif isinstance(prior_hyper_parameters, dict):
            prior_hyper_parameters = [prior_hyper_parameters]
        self.prior_hyper_parameters = prior_hyper_parameters

    @abstractmethod
    def print_result(self):
        """Print a HPO algorithm result."""
        raise NotImplementedError

    @abstractmethod
    def save_results(self):
        """Save a HPO algorithm result."""
        raise NotImplementedError

    @abstractmethod
    def is_done(self):
        """Check whether HPO algorithm is done."""
        raise NotImplementedError

    @abstractmethod
    def get_next_sample(self):
        """Get next sample to train."""
        raise NotImplementedError

    @abstractmethod
    def auto_config(self):
        """Configure HPO algorithm automatically."""
        raise NotImplementedError

    @abstractmethod
    def get_progress(self):
        """Get current progress of HPO algorithm."""
        raise NotImplementedError

    @abstractmethod
    def report_score(self, score, resource, trial_id, done):
        """Report a score to HPO algorithm."""
        raise NotImplementedError

    @abstractmethod
    def get_best_config(self):
        """Get best config of HPO algorithm."""
        raise NotImplementedError


class Trial:
    """Trial to train with given hyper parameters.

    Args:
        trial_id (Any): Trial id.
        configuration (Dict): Configuration to train with.
        train_environment (Optional[Dict], optional): Train environment for the trial. Defaults to None.
    """

    def __init__(self, trial_id: Any, configuration: Dict, train_environment: Optional[Dict] = None):
        self._id = trial_id
        self._configuration = configuration
        self.score: Dict[Union[float, int], Union[float, int]] = {}
        self._train_environment = train_environment
        self._iteration = None
        self.status: TrialStatus = TrialStatus.READY

    @property
    def id(self):
        """Trial id."""
        return self._id

    @property
    def configuration(self):
        """Configuration to train with."""
        return self._configuration

    @property
    def iteration(self):
        """Iteration to use for training."""
        return self._iteration

    @iteration.setter
    def iteration(self, val):
        """Setter for iteration."""
        check_positive(val, "iteration")
        self._iteration = val

    @property
    def train_environment(self):
        """Train environment for the trial."""
        return self._train_environment

    def get_train_configuration(self) -> Dict[str, Any]:
        """Get configurations needed to trian."""
        self._configuration["iterations"] = self.iteration
        return {"id": self.id, "configuration": self.configuration, "train_environment": self.train_environment}

    def register_score(self, score: Union[int, float], resource: Union[int, float]):
        """Register score to the trial.

        Args:
            score (Union[int, float]): Score to register.
            resource (Union[int, float]): Resource used to get score. It should be positive.
        """
        check_positive(resource, "resource")
        self.score[resource] = score

    def get_best_score(
        self, mode: str = "max", resource_limit: Optional[Union[float, int]] = None
    ) -> Optional[Union[float, int]]:
        """Get best score of the trial.

        Args:
            mode (str, optional): Decide which is better between highest score or lowest score. Defaults to "max".
            resource_limit (Optional[Union[float, int]], optional): Find a best score among the score at resource
                                                                    lower than this value. Defaults to None.

        Returns:
            Optional[Union[float, int]]: Best score. If there is no score, return None.
        """
        check_mode_input(mode)

        if resource_limit is None:
            scores = self.score.values()
        else:
            scores = [val for key, val in self.score.items() if key <= resource_limit]  # type: ignore

        if len(scores) == 0:
            return None

        if mode == "max":
            return max(scores)
        return min(scores)

    def get_progress(self) -> Union[float, int]:
        """Get a progress of the trial.

        Returns:
            Union[float, int]: How many resource is used for the trial.
        """
        if len(self.score) == 0:
            return 0
        return max(self.score.keys())

    def save_results(self, save_path: str):
        """Save a result in the 'save_path'.

        Args:
            save_path (str): Path where to save a result.
        """
        results = {
            "id": self.id,
            "configuration": self.configuration,
            "train_environment": self.train_environment,
            "score": self.score,
        }

        with open(save_path, "w", encoding="utf-8") as f:
            json.dump(results, f)

    def finalize(self):
        """Let the trial know that training is done.

        If the trial isn't trained until given resource, then make it pretend to train until resouce.
        """
        if self.get_progress() < self.iteration:
            best_score = self.get_best_score()
            if best_score is None:
                raise RuntimeError(f"Although {self.id} trial doesn't report any score but it's done")
            self.register_score(best_score, self.iteration)

    def is_done(self):
        """Check the trial is done."""
        if self.iteration is None:
            raise ValueError("iteration isn't set yet.")
        return self.get_progress() >= self.iteration


class TrialStatus(IntEnum):
    """Enum class for trial status."""

    UNKNOWN = -1
    READY = 0
    RUNNING = 1
    STOP = 2
    CUDAOOM = 3



import mmcv

from otx.mpa.stage import Stage
from otx.mpa.workflow import Workflow
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestWorkflow:
    @e2e_pytest_unit
    def test_run(self, mocker):
        fake_cfg = {
            "work_dir": "/path/workdir",
            "seed": 0,
        }
        fake_common_cfg = {"output_path": "/path/output"}
        mocker.patch.object(mmcv, "mkdir_or_exist")
        stage = Stage(
            "MockStage",
            "",
            fake_cfg,
            fake_common_cfg,
            0,
            input=dict(
                arg_name1=dict(stage_name="MockStage", output_key=""),
                arg_name2=dict(stage_name="MockStage", output_key=""),
            ),
        )
        workflow = Workflow([stage])

        mock_stage_run = mocker.patch.object(stage, "run")
        workflow.run()

        mock_stage_run.assert_called()



"""Base Dataset for Classification Task."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# pylint: disable=invalid-name, too-many-locals, no-member

from typing import Any, Dict, List

import numpy as np
from mmcls.core import average_performance, mAP
from mmcls.datasets.base_dataset import BaseDataset
from mmcls.datasets.builder import DATASETS, PIPELINES
from mmcls.datasets.pipelines import Compose
from mmcv.utils.registry import build_from_cfg
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix
from torch.utils.data import Dataset

from otx.algorithms.common.utils import get_cls_img_indices, get_old_new_img_indices
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()


# pylint: disable=too-many-instance-attributes
@DATASETS.register_module()
class OTXClsDataset(BaseDataset):
    """Multi-class classification dataset class."""

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    def __init__(
        self, otx_dataset: DatasetEntity, labels: List[LabelEntity], empty_label=None, **kwargs
    ):  # pylint: disable=super-init-not-called
        self.otx_dataset = otx_dataset
        self.labels = labels
        self.label_names = [label.name for label in self.labels]
        self.label_idx = {label.id: i for i, label in enumerate(labels)}
        self.empty_label = empty_label
        self.class_acc = False

        self.CLASSES = list(label.name for label in labels)
        self.gt_labels = []  # type: List
        pipeline = kwargs.get("pipeline", [])
        self.num_classes = len(self.CLASSES)

        test_mode = kwargs.get("test_mode", False)
        if test_mode is False:
            new_classes = kwargs.pop("new_classes", [])
            self.img_indices = self.get_indices(new_classes)

        _pipeline = [dict(type="LoadImageFromOTXDataset"), *pipeline]
        self.pipeline = Compose([build_from_cfg(p, PIPELINES) for p in _pipeline])
        self.load_annotations()

    def get_indices(self, *args):  # pylint: disable=unused-argument
        """Get indices."""
        return get_cls_img_indices(self.labels, self.otx_dataset)

    def load_annotations(self):
        """Load annotations."""
        include_empty = self.empty_label in self.labels
        for i, _ in enumerate(self.otx_dataset):
            class_indices = []
            item_labels = self.otx_dataset[i].get_roi_labels(self.labels, include_empty=include_empty)
            ignored_labels = self.otx_dataset[i].ignored_labels
            if item_labels:
                for otx_lbl in item_labels:
                    if otx_lbl not in ignored_labels:
                        class_indices.append(self.label_names.index(otx_lbl.name))
                    else:
                        class_indices.append(-1)
            else:  # this supposed to happen only on inference stage
                class_indices.append(-1)
            self.gt_labels.append(class_indices)
        self.gt_labels = np.array(self.gt_labels)

    @check_input_parameters_type()
    def __getitem__(self, index: int):
        """Get item from dataset."""
        dataset = self.otx_dataset
        item = dataset[index]
        ignored_labels = np.array([self.label_idx[lbs.id] for lbs in item.ignored_labels])

        height, width = item.height, item.width

        data_info = dict(
            dataset_item=item,
            width=width,
            height=height,
            index=index,
            gt_label=self.gt_labels[index],
            ignored_labels=ignored_labels,
        )

        if self.pipeline is None:
            return data_info
        return self.pipeline(data_info)

    def get_gt_labels(self):
        """Get all ground-truth labels (categories).

        Returns:
            list[int]: categories for all images.
        """

        return self.gt_labels

    def __len__(self):
        """Get dataset length."""
        return len(self.otx_dataset)

    def evaluate(
        self, results, metric="accuracy", metric_options=None, logger=None
    ):  # pylint: disable=redefined-outer-name
        """Evaluate the dataset with new metric class_accuracy.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
                Default value is `accuracy`.
                'accuracy', 'precision', 'recall', 'f1_score', 'support', 'class_accuracy'
            metric_options (dict, optional): Options for calculating metrics.
                Allowed keys are 'topk', 'thrs' and 'average_mode'.
                Defaults to None.
            logger (logging.Logger | str, optional): Logger used for printing
                related information during evaluation. Defaults to None.

        Returns:
            dict: evaluation results
        """

        if metric_options is None:
            metric_options = {"topk": (1, 5) if self.num_classes >= 5 else (1,)}

        if isinstance(metric, str):
            metrics = [metric]
        else:
            metrics = metric

        if "class_accuracy" in metrics:
            metrics.remove("class_accuracy")
            self.class_acc = True

        eval_results = super().evaluate(results, metrics, metric_options, logger=logger)
        for k in metric_options["topk"]:
            eval_results[f"accuracy_top-{k}"] /= 100

        # Add Evaluation Accuracy score per Class - it can be used only for multi-class dataset.
        if self.class_acc:
            results = np.vstack(results)
            gt_labels = self.get_gt_labels()
            accuracies = self.class_accuracy(results, gt_labels)

            if any(np.isnan(accuracies)):
                accuracies = np.nan_to_num(accuracies)

            eval_results.update({f"{c} accuracy": a for c, a in zip(self.CLASSES, accuracies)})
            eval_results.update({"mean accuracy": np.mean(accuracies)})

        eval_results["accuracy"] = eval_results["accuracy_top-1"]
        return eval_results

    def class_accuracy(self, results, gt_labels):
        """Return per-class accuracy."""
        accracies = []
        pred_label = results.argsort(axis=1)[:, -1:][:, ::-1]
        for i in range(self.num_classes):
            cls_pred = pred_label == i
            cls_pred = cls_pred[gt_labels == i]
            cls_acc = np.sum(cls_pred) / len(cls_pred)
            accracies.append(cls_acc)
        return accracies


@DATASETS.register_module()
class OTXMultilabelClsDataset(OTXClsDataset):
    """Multi-label classification dataset class."""

    def get_indices(self, new_classes):
        """Get indices."""
        return get_old_new_img_indices(self.labels, new_classes, self.otx_dataset)

    def load_annotations(self):
        """Load annotations."""
        include_empty = self.empty_label in self.labels
        for i, _ in enumerate(self.otx_dataset):
            item_labels = self.otx_dataset[i].get_roi_labels(self.labels, include_empty=include_empty)
            ignored_labels = self.otx_dataset[i].ignored_labels
            onehot_indices = np.zeros(len(self.labels))
            if item_labels:
                for otx_lbl in item_labels:
                    if otx_lbl not in ignored_labels:
                        onehot_indices[self.label_names.index(otx_lbl.name)] = 1
                    else:
                        # during training we filter ignored classes out,
                        # during validation mmcv's mAP also filters -1 labels
                        onehot_indices[self.label_names.index(otx_lbl.name)] = -1

            self.gt_labels.append(onehot_indices)
        self.gt_labels = np.array(self.gt_labels)

    def evaluate(
        self, results, metric="mAP", metric_options=None, indices=None, logger=None
    ):  # pylint: disable=unused-argument, redefined-outer-name, arguments-renamed
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
                Default value is 'mAP'. Options are 'mAP', 'CP', 'CR', 'CF1',
                'OP', 'OR' and 'OF1'.
            metric_options (dict, optional): Options for calculating metrics.
                Allowed keys are 'k' and 'thr'. Defaults to None
            logger (logging.Logger | str, optional): Logger used for printing
                related information during evaluation. Defaults to None.

        Returns:
            dict: evaluation results
        """
        if metric_options is None or metric_options == {}:
            metric_options = {"thr": 0.5}

        if isinstance(metric, str):
            metrics = [metric]
        else:
            metrics = metric
        allowed_metrics = ["accuracy-mlc", "mAP", "CP", "CR", "CF1", "OP", "OR", "OF1"]
        eval_results = {}
        results = np.vstack(results)
        gt_labels = self.get_gt_labels()
        if indices is not None:
            gt_labels = gt_labels[indices]
        num_imgs = len(results)
        assert len(gt_labels) == num_imgs, "dataset testing results should " "be of the same length as gt_labels."

        invalid_metrics = set(metrics) - set(allowed_metrics)
        if len(invalid_metrics) != 0:
            raise ValueError(f"metric {invalid_metrics} is not supported.")

        if "accuracy-mlc" in metrics:
            true_label_idx = []
            pred_label_idx = []
            pos_thr = metric_options.get("thr", 0.5)

            true_label = gt_labels == 1
            pred_label = results > pos_thr
            cls_index = [i + 1 for i in range(len(self.labels))]
            for true_lbl, pred_lbl in zip(true_label, pred_label):
                true_lbl_idx = set(true_lbl * cls_index) - set([0])  # except empty
                pred_lbl_idx = set(pred_lbl * cls_index) - set([0])
                true_label_idx.append(true_lbl_idx)
                pred_label_idx.append(pred_lbl_idx)

            confusion_matrices = []
            for cls_idx in cls_index:
                group_labels_idx = set([cls_idx - 1])
                y_true = [int(not group_labels_idx.issubset(true_labels)) for true_labels in true_label_idx]
                y_pred = [int(not group_labels_idx.issubset(pred_labels)) for pred_labels in pred_label_idx]
                matrix_data = sklearn_confusion_matrix(y_true, y_pred, labels=list(range(len([0, 1]))))
                confusion_matrices.append(matrix_data)
            correct_per_label_group = [np.trace(mat) for mat in confusion_matrices]
            total_per_label_group = [np.sum(mat) for mat in confusion_matrices]

            acc = np.sum(correct_per_label_group) / np.sum(total_per_label_group)  # MICRO average
            eval_results["accuracy-mlc"] = acc
            eval_results["accuracy"] = eval_results["accuracy-mlc"]

        if "mAP" in metrics:
            mAP_value = mAP(results, gt_labels)
            eval_results["mAP"] = mAP_value
        if len(set(metrics) - {"mAP"}) != 0:
            performance_keys = ["CP", "CR", "CF1", "OP", "OR", "OF1"]
            performance_values = average_performance(results, gt_labels, **metric_options)
            for k, v in zip(performance_keys, performance_values):
                if k in metrics:
                    eval_results[k] = v

        return eval_results


@DATASETS.register_module()
class OTXHierarchicalClsDataset(OTXMultilabelClsDataset):
    """Hierarchical classification dataset class."""

    def __init__(self, **kwargs):
        self.hierarchical_info = kwargs.pop("hierarchical_info", None)
        super().__init__(**kwargs)

    def load_annotations(self):
        """Load annotations."""
        include_empty = self.empty_label in self.labels
        for i, _ in enumerate(self.otx_dataset):
            class_indices = []
            item_labels = self.otx_dataset[i].get_roi_labels(self.labels, include_empty=include_empty)
            ignored_labels = self.otx_dataset[i].ignored_labels
            if item_labels:
                num_cls_heads = self.hierarchical_info["num_multiclass_heads"]

                class_indices = [0] * (
                    self.hierarchical_info["num_multiclass_heads"] + self.hierarchical_info["num_multilabel_classes"]
                )
                for j in range(num_cls_heads):
                    class_indices[j] = -1
                for otx_lbl in item_labels:
                    group_idx, in_group_idx = self.hierarchical_info["class_to_group_idx"][otx_lbl.name]
                    if group_idx < num_cls_heads:
                        class_indices[group_idx] = in_group_idx
                    elif otx_lbl not in ignored_labels:
                        class_indices[num_cls_heads + in_group_idx] = 1
                    else:
                        class_indices[num_cls_heads + in_group_idx] = -1
            else:  # this supposed to happen only on inference stage or if we have a negative in multilabel data
                class_indices = [-1] * (
                    self.hierarchical_info["num_multiclass_heads"] + self.hierarchical_info["num_multilabel_classes"]
                )
            self.gt_labels.append(class_indices)
        self.gt_labels = np.array(self.gt_labels)

    @staticmethod
    def mean_top_k_accuracy(scores, labels, k=1):
        """Return mean of top-k accuracy."""
        idx = np.argsort(-scores, axis=-1)[:, :k]
        labels = np.array(labels)
        matches = np.any(idx == labels.reshape([-1, 1]), axis=-1)

        classes = np.unique(labels)

        accuracy_values = []
        for class_id in classes:
            mask = labels == class_id
            num_valid = np.sum(mask)
            if num_valid == 0:
                continue

            accuracy_values.append(np.sum(matches[mask]) / float(num_valid))

        return np.mean(accuracy_values) * 100 if len(accuracy_values) > 0 else 1.0

    def evaluate(
        self, results, metric="MHAcc", metric_options=None, indices=None, logger=None
    ):  # pylint: disable=unused-argument, redefined-outer-name
        """Evaluate the dataset.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
                Default value is 'mAP'. Options are 'mAP', 'CP', 'CR', 'CF1',
                'OP', 'OR' and 'OF1'.
            metric_options (dict, optional): Options for calculating metrics.
                Allowed keys are 'k' and 'thr'. Defaults to None
            logger (logging.Logger | str, optional): Logger used for printing
                related information during evaluation. Defaults to None.

        Returns:
            dict: evaluation results
        """
        if metric_options is None or metric_options == {}:
            metric_options = {"thr": 0.5}

        if isinstance(metric, str):
            metrics = [metric]
        else:
            metrics = metric

        allowed_metrics = ["MHAcc", "avgClsAcc", "mAP"]
        eval_results = {}
        results = np.vstack(results)
        gt_labels = self.get_gt_labels()
        if indices is not None:
            gt_labels = gt_labels[indices]
        num_imgs = len(results)
        assert len(gt_labels) == num_imgs, "dataset testing results should " "be of the same length as gt_labels."

        invalid_metrics = set(metrics) - set(allowed_metrics)
        if len(invalid_metrics) != 0:
            raise ValueError(f"metric {invalid_metrics} is not supported.")

        total_acc = 0.0
        total_acc_sl = 0.0
        for i in range(self.hierarchical_info["num_multiclass_heads"]):
            multiclass_logit = results[
                :,
                self.hierarchical_info["head_idx_to_logits_range"][i][0] : self.hierarchical_info[
                    "head_idx_to_logits_range"
                ][i][1],
            ]  # noqa: E127
            multiclass_gt = gt_labels[:, i]
            cls_acc = self.mean_top_k_accuracy(multiclass_logit, multiclass_gt, k=1)
            total_acc += cls_acc
            total_acc_sl += cls_acc

        mAP_value = 0.0
        if self.hierarchical_info["num_multilabel_classes"] and "mAP" in metrics:
            multilabel_logits = results[:, self.hierarchical_info["num_single_label_classes"] :]
            multilabel_gt = gt_labels[:, self.hierarchical_info["num_multiclass_heads"] :]
            mAP_value = mAP(multilabel_logits, multilabel_gt)

        total_acc += mAP_value
        total_acc /= self.hierarchical_info["num_multiclass_heads"] + int(
            self.hierarchical_info["num_multilabel_classes"] > 0
        )

        eval_results["MHAcc"] = total_acc
        eval_results["avgClsAcc"] = total_acc_sl / self.hierarchical_info["num_multiclass_heads"]
        eval_results["mAP"] = mAP_value
        eval_results["accuracy"] = total_acc

        return eval_results


@DATASETS.register_module()
class SelfSLDataset(Dataset):
    """SelfSL dataset that training with two pipelines and no label."""

    CLASSES = None

    @check_input_parameters_type({"otx_dataset": DatasetParamTypeCheck})
    def __init__(
        self, otx_dataset: DatasetEntity, pipeline: Dict[str, Any], **kwargs
    ):  # pylint: disable=unused-argument
        super().__init__()
        self.otx_dataset = otx_dataset

        self.load_pipeline = build_from_cfg(dict(type="LoadImageFromOTXDataset"), PIPELINES)
        self.view0 = Compose([build_from_cfg(p, PIPELINES) for p in pipeline["view0"]])
        self.view1 = Compose([build_from_cfg(p, PIPELINES) for p in pipeline["view1"]])

    def __len__(self):
        """Get dataset length."""
        return len(self.otx_dataset)

    def __getitem__(self, index: int):
        """Get item from dataset."""
        dataset = self.otx_dataset
        item = dataset[index]

        height, width = item.height, item.width

        data_info = dict(
            dataset_item=item,
            width=width,
            height=height,
            index=index,
        )

        loaded_results = self.load_pipeline(data_info)
        results1 = self.view0(loaded_results.copy())
        results2 = self.view1(loaded_results.copy())

        results = {}
        for k, v in results1.items():
            results[k + "1"] = v
        for k, v in results2.items():
            results[k + "2"] = v

        return results



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os

MPA_CACHE = os.path.expanduser(os.getenv("MPA_CACHE", os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "mpa")))
os.makedirs(MPA_CACHE, exist_ok=True)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""This module define the scored label entity."""

import datetime
import math
from dataclasses import dataclass
from typing import Optional

from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity


@dataclass
class LabelSource:
    """This dataclass contains information about the source of a scored label.

    For annotations, the id of the user who created the label and for predictions, the
    id and model storage id of the model that created the prediction. When a user has
    accepted a predictions as is, both the user id of the user who accepted and the
    model/model storage id of the model that predicted should be filled in.
    """

    user_id: str = ""
    model_id: ID = ID()
    model_storage_id: ID = ID()


class ScoredLabel:
    """This represents a label along with a probability. This is used inside `Annotation` class.

    Args:
        label (LabelEntity): Label entity to which probability and source are attached.
        probability (float): a float denoting the probability of the shape belonging to the label.
        label_source (LabelSource): a LabelSource dataclass containing the id of the user who created
            or the model that predicted this label.
    """

    def __init__(
        self,
        label: LabelEntity,
        probability: float = 0.0,
        label_source: Optional[LabelSource] = None,
    ):
        if math.isnan(probability) or (not 0 <= probability <= 1.0):
            raise ValueError(f"Probability should be in range [0, 1], {probability} is given")

        self.label = label
        self.probability = probability
        self.label_source = label_source if label_source is not None else LabelSource()

    @property
    def name(self) -> str:
        """Name of the label."""
        return self.label.name

    @property
    def id_(self) -> ID:
        """Returns the label id."""
        return self.label.id_

    @property
    def id(self) -> ID:
        """DEPRECATED."""
        return self.label.id

    @property
    def color(self) -> Color:
        """Color of the label."""
        return self.label.color

    @property
    def hotkey(self) -> str:
        """Hotkey of the label."""
        return self.label.hotkey

    @property
    def domain(self) -> Domain:
        """Domain of the label."""
        return self.label.domain

    @property
    def is_empty(self) -> bool:
        """Check if the label is empty."""
        return self.label.is_empty

    @property
    def creation_date(self) -> datetime.datetime:
        """Creation data of the label."""
        return self.label.creation_date

    def get_label(self) -> LabelEntity:
        """Gets the label that the ScoredLabel object was initialized with."""
        return self.label

    def __repr__(self):
        """String representation of the label."""
        return (
            f"ScoredLabel({self.id_}, name={self.name}, probability={self.probability}, "
            f"domain={self.domain}, color={self.color}, hotkey={self.hotkey}, "
            f"label_source={self.label_source})"
        )

    def __eq__(self, other: object) -> bool:
        """Checks if the label is equal to the other label.

        Args:
            other (ScoredLabel): Label to compare with

        Returns:
            bool: True if the labels are equal, False otherwise
        """
        if isinstance(other, ScoredLabel):
            return (
                self.id_ == other.id_
                and self.name == other.name
                and self.color == other.color
                and self.hotkey == other.hotkey
                and self.probability == other.probability
                and self.domain == other.domain
                and self.label_source == other.label_source
            )
        return False

    def __hash__(self):
        """Returns hash of the label."""
        return hash(str(self))



"""This module contains the implementation of Accuracy performance provider."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


import copy
import logging
from typing import List, Set, Tuple

import numpy as np
from sklearn.metrics import confusion_matrix as sklearn_confusion_matrix

from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.label_schema import LabelGroup
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    ColorPalette,
    MatrixChartInfo,
    MatrixMetric,
    MatrixMetricsGroup,
    MetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.averaging import MetricAverageMethod
from otx.api.usecases.evaluation.basic_operations import (
    precision_per_class,
    recall_per_class,
)
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)

logger = logging.getLogger(__name__)


class Accuracy(IPerformanceProvider):
    """This class is responsible for providing Accuracy measures; mainly for Classification problems.

    The calculation both supports multi label and binary label predictions.

    Accuracy is the proportion of the predicted correct labels, to the total number (predicted and actual)
    labels for that instance. Overall accuracy is the average across all instances.

    Args:
        resultset (ResultSetEntity): ResultSet that score will be computed for
        average (MetricAverageMethod, optional): The averaging method, either MICRO or MACRO
            MICRO: compute average over all predictions in all label groups
            MACRO: compute accuracy per label group, return the average of the per-label-group accuracy scores
    """

    def __init__(
        self,
        resultset: ResultSetEntity,
        average: MetricAverageMethod = MetricAverageMethod.MICRO,
    ):
        self._unnormalized_matrices: List[MatrixMetric] = compute_unnormalized_confusion_matrices_from_resultset(
            resultset
        )

        # accuracy computation
        mean_accuracy = self._compute_accuracy(average=average, confusion_matrices=self._unnormalized_matrices)
        self._accuracy = ScoreMetric(value=mean_accuracy, name="Accuracy")

    @property
    def accuracy(self) -> ScoreMetric:
        """Returns the accuracy as ScoreMetric."""
        return self._accuracy

    def get_performance(self) -> Performance:
        """Returns the performance with accuracy and confusion metrics."""
        confusion_matrix_dashboard_metrics: List[MetricsGroup] = []

        # Use normalized matrix for UI
        normalized_matrices: List[MatrixMetric] = copy.deepcopy(self._unnormalized_matrices)
        for unnormalized_matrix in normalized_matrices:
            unnormalized_matrix.normalize()

        confusion_matrix_info = MatrixChartInfo(
            name="Confusion matrix",
            header="confusion",
            row_header="Predicted label",
            column_header="True label",
        )
        confusion_matrix_dashboard_metrics.append(
            MatrixMetricsGroup(metrics=normalized_matrices, visualization_info=confusion_matrix_info)
        )
        #  Compute precision and recall MetricGroups and append them to the dashboard metrics
        for _confusion_matrix in self._unnormalized_matrices:
            confusion_matrix_dashboard_metrics.append(precision_metrics_group(_confusion_matrix))
            confusion_matrix_dashboard_metrics.append(recall_metrics_group(_confusion_matrix))

        return Performance(score=self.accuracy, dashboard_metrics=confusion_matrix_dashboard_metrics)

    @staticmethod
    def _compute_accuracy(average: MetricAverageMethod, confusion_matrices: List[MatrixMetric]) -> float:
        """Compute accuracy using the confusion matrices.

        Args:
            average (MatricAverageMethod): The averaging method, either MICRO or MACRO
                MICRO: compute average over all predictions in all label groups
                MACRO: compute accuracy per label group, return the average of the per-label-group accuracy scores
            confusion_matrices (List[MatrixMetric]): the confusion matrices to compute accuracy from.
                MUST be unnormalized.

        Raises
            ValueError: when the ground truth dataset does not contain annotations
            RuntimeError: when the averaging methods is not known
        Returns:
            float: the accuracy score for the provided confusion matrix
        """
        # count correct predictions and total annotations
        correct_per_label_group = [np.trace(mat.matrix_values) for mat in confusion_matrices]
        total_per_label_group = [np.sum(mat.matrix_values) for mat in confusion_matrices]
        # check if all label groups have annotations
        if not np.any(total_per_label_group):
            raise ValueError("The ground truth dataset must contain annotations.")
        # return micro or macro average
        if average == MetricAverageMethod.MACRO:
            # compute accuracy for each label group, then average across groups, ignoring groups without annotations
            return np.nanmean(np.divide(correct_per_label_group, total_per_label_group))
        if average == MetricAverageMethod.MICRO:
            # average over all predictions in all label groups
            return np.sum(correct_per_label_group) / np.sum(total_per_label_group)

        raise RuntimeError(f"Unknown averaging method: {average}")


def precision_metrics_group(confusion_matrix: MatrixMetric) -> MetricsGroup:
    """Computes the precision per class based on a confusion matrix and returns them as ScoreMetrics in a MetricsGroup.

    Args:
        confusion_matrix: matrix to compute the precision per class for

    Returns:
        a BarMetricsGroup with the per class precision.
    """
    labels = confusion_matrix.row_labels
    if labels is None:
        # If no labels are given, just number the classes by index
        if confusion_matrix.matrix_values is not None:
            label_range = confusion_matrix.matrix_values.shape[0]
        else:
            label_range = 0
        labels = np.arange(label_range)

    per_class_precision = [
        ScoreMetric(class_, value=precision)
        for (class_, precision) in zip(labels, precision_per_class(confusion_matrix.matrix_values))
    ]

    return BarMetricsGroup(
        metrics=per_class_precision,
        visualization_info=BarChartInfo(
            name="Precision per class",
            palette=ColorPalette.LABEL,
        ),
    )


def recall_metrics_group(confusion_matrix: MatrixMetric) -> MetricsGroup:
    """Computes the recall per class based on a confusion matrix and returns them as ScoreMetrics in a MetricsGroup.

    Args:
        confusion_matrix: matrix to compute the recall per class for

    Returns:
        a BarMetricsGroup with the per class recall
    """
    labels = confusion_matrix.row_labels
    if labels is None:
        # If no labels are given, just number the classes by index
        if confusion_matrix.matrix_values is not None:
            label_range = confusion_matrix.matrix_values.shape[0]
        else:
            label_range = 0
        labels = np.arange(label_range)

    per_class_recall = [
        ScoreMetric(class_, value=recall)
        for (class_, recall) in zip(labels, recall_per_class(confusion_matrix.matrix_values))
    ]

    return BarMetricsGroup(
        metrics=per_class_recall,
        visualization_info=BarChartInfo(
            name="Recall per class",
            palette=ColorPalette.LABEL,
        ),
    )


def __get_gt_and_predicted_label_indices_from_resultset(
    resultset: ResultSetEntity,
) -> Tuple[List[Set[int]], List[Set[int]]]:
    """Returns the label indices lists for ground truth and prediction datasets in a tuple.

    Args:
        resultset

    Returns:
        a tuple containing two lists. The first list contains the ground truth label indices, and the second contains
        the prediction label indices.
    """
    true_label_idx = []
    predicted_label_idx = []

    gt_dataset: DatasetEntity = resultset.ground_truth_dataset
    pred_dataset: DatasetEntity = resultset.prediction_dataset

    gt_dataset.sort_items()
    pred_dataset.sort_items()

    # Iterate over each dataset item, and collect the labels for this item (pred and gt)
    task_labels = resultset.model.configuration.get_label_schema().get_labels(include_empty=True)
    for gt_item, pred_item in zip(gt_dataset, pred_dataset):
        if isinstance(gt_item, DatasetItemEntity) and isinstance(pred_item, DatasetItemEntity):
            true_label_idx.append({task_labels.index(label) for label in gt_item.get_roi_labels(task_labels)})
            predicted_label_idx.append({task_labels.index(label) for label in pred_item.get_roi_labels(task_labels)})

    return true_label_idx, predicted_label_idx


def __compute_unnormalized_confusion_matrices_for_label_group(
    true_label_idx: List[Set[int]],
    predicted_label_idx: List[Set[int]],
    label_group: LabelGroup,
    task_labels: List[LabelEntity],
) -> MatrixMetric:
    """Returns matrix metric for a certain label group.

    Args:
        true_label_idx (List[Set[int]]): list of sets of label indices for the ground truth dataset
        predicted_label_idx (List[Set[int]]): list of sets of label indices for the prediction dataset
        label_group (LabelGroup): label group to compute the confusion matrix for
        task_labels (List[LabelEntity]): list of labels for the task

    Returns:
        MatrixMetric: confusion matrix for the label group
    """
    map_task_labels_idx_to_group_idx = {
        task_labels.index(label): i_group for i_group, label in enumerate(label_group.labels)
    }
    set_group_labels_idx = set(map_task_labels_idx_to_group_idx.keys())
    group_label_names = [task_labels[label_idx].name for label_idx in set_group_labels_idx]

    if len(group_label_names) == 1:
        # Single-class
        # we use "not" to make presence of a class to be at index 0, while the absence of it at index 1
        y_true = [int(not set_group_labels_idx.issubset(true_labels)) for true_labels in true_label_idx]
        y_pred = [int(not set_group_labels_idx.issubset(pred_labels)) for pred_labels in predicted_label_idx]
        group_label_names += [f"~ {group_label_names[0]}"]
        column_labels = group_label_names.copy()
        remove_last_row = False
    else:
        # Multiclass
        undefined_idx = len(group_label_names)  # to define missing value

        # find the intersections between GT and task labels, and Prediction and task labels
        true_intersections = [true_labels.intersection(set_group_labels_idx) for true_labels in true_label_idx]
        pred_intersections = [pred_labels.intersection(set_group_labels_idx) for pred_labels in predicted_label_idx]

        # map the intersection to 0-index value
        y_true = [
            map_task_labels_idx_to_group_idx[list(true_intersection)[0]]
            if len(true_intersection) != 0
            else undefined_idx
            for true_intersection in true_intersections
        ]
        y_pred = [
            map_task_labels_idx_to_group_idx[list(pred_intersection)[0]]
            if len(pred_intersection) != 0
            else undefined_idx
            for pred_intersection in pred_intersections
        ]

        column_labels = group_label_names.copy()
        column_labels.append("Other")
        remove_last_row = True

    matrix_data = sklearn_confusion_matrix(y_true, y_pred, labels=list(range(len(column_labels))))
    if remove_last_row:
        # matrix clean up
        matrix_data = np.delete(matrix_data, -1, 0)
        if sum(matrix_data[:, -1]) == 0:
            # if none of the GT is classified as classes from other groups, clean it up too
            matrix_data = np.delete(matrix_data, -1, 1)
            column_labels.remove(column_labels[-1])

    # Use unnormalized matrix for statistics computation (accuracy, precision, recall)
    return MatrixMetric(
        name=f"{label_group.name}",
        matrix_values=matrix_data,
        row_labels=group_label_names,
        column_labels=column_labels,
        normalize=False,
    )


def compute_unnormalized_confusion_matrices_from_resultset(
    resultset: ResultSetEntity,
) -> List[MatrixMetric]:
    """Computes an (unnormalized) confusion matrix for every label group in the resultset.

    Args:
        resultset: the input resultset

    Returns:
        the computed unnormalized confusion matrices
    """

    if len(resultset.ground_truth_dataset) == 0 or len(resultset.prediction_dataset) == 0:
        raise ValueError("Cannot compute the confusion matrix of an empty result set.")

    unnormalized_confusion_matrices: List[MatrixMetric] = []
    (
        true_label_idx,
        predicted_label_idx,
    ) = __get_gt_and_predicted_label_indices_from_resultset(resultset)
    task_labels = resultset.model.configuration.get_label_schema().get_labels(include_empty=False)

    # Confusion matrix computation
    for label_group in resultset.model.configuration.get_label_schema().get_groups():
        matrix = __compute_unnormalized_confusion_matrices_for_label_group(
            true_label_idx, predicted_label_idx, label_group, task_labels
        )
        unnormalized_confusion_matrices.append(matrix)

    return unnormalized_confusion_matrices



"""This module contains the Dice performance provider."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import Dict, List, Optional, Tuple

from otx.api.entities.label import LabelEntity
from otx.api.entities.metrics import (
    BarChartInfo,
    BarMetricsGroup,
    ColorPalette,
    MetricsGroup,
    Performance,
    ScoreMetric,
)
from otx.api.entities.resultset import ResultSetEntity
from otx.api.usecases.evaluation.averaging import MetricAverageMethod
from otx.api.usecases.evaluation.basic_operations import (
    get_intersections_and_cardinalities,
)
from otx.api.usecases.evaluation.performance_provider_interface import (
    IPerformanceProvider,
)
from otx.api.utils.segmentation_utils import mask_from_dataset_item
from otx.api.utils.time_utils import timeit


class DiceAverage(IPerformanceProvider):
    """Computes the average Dice coefficient overall and for individual labels.

    See https://en.wikipedia.org/wiki/S%C3%B8rensen%E2%80%93Dice_coefficient for background information.

    To compute the Dice coefficient the shapes in the dataset items of the prediction and ground truth
    dataset are first converted to masks.

    Dice is computed by computing the intersection and union computed over the whole dataset, instead of
    computing intersection and union for individual images and then averaging.

    Args:
        resultset (ResultSetEntity): ResultSet that score will be computed for
        average (MetricAverageMethod): One of
            - MICRO: every pixel has the same weight, regardless of label
            - MACRO: compute score per label, return the average of the per-label scores
    """

    def __init__(
        self,
        resultset: ResultSetEntity,
        average: MetricAverageMethod = MetricAverageMethod.MACRO,
    ):
        self.average = average
        (
            self._overall_dice,
            self._dice_per_label,
        ) = self.__compute_dice_averaged_over_pixels(resultset, average)

    @property
    def overall_dice(self) -> ScoreMetric:
        """Returns the dice average as ScoreMetric."""
        return self._overall_dice

    @property
    def dice_per_label(self) -> Dict[LabelEntity, ScoreMetric]:
        """Returns a dictionary mapping the label to its corresponding dice score (as ScoreMetric)."""
        return self._dice_per_label

    def get_performance(self) -> Performance:
        """Returns the performance of the resultset."""
        score = self.overall_dice
        dashboard_metrics: Optional[List[MetricsGroup]]
        if len(self.dice_per_label) == 0:
            dashboard_metrics = None
        else:
            dashboard_metrics = [
                BarMetricsGroup(
                    metrics=list(self.dice_per_label.values()),
                    visualization_info=BarChartInfo(
                        name="Dice Average Per Label",
                        palette=ColorPalette.LABEL,
                    ),
                )
            ]
        return Performance(score=score, dashboard_metrics=dashboard_metrics)

    @classmethod
    @timeit
    def __compute_dice_averaged_over_pixels(
        cls, resultset: ResultSetEntity, average: MetricAverageMethod
    ) -> Tuple[ScoreMetric, Dict[LabelEntity, ScoreMetric]]:
        """Computes the diced averaged over pixels.

        Args:
            resultset (ResultSetEntity): Result set to use
            average (MetricAverageMethod): Averaging method to use

        Returns:
            Tuple[ScoreMetric, Dict[LabelEntity, ScoreMetric]]: Tuple of the overall dice and the dice averaged over
                pixels for each label.
        """
        if len(resultset.prediction_dataset) == 0:
            raise ValueError("Cannot compute the DICE score of an empty result set.")

        if len(resultset.prediction_dataset) != len(resultset.ground_truth_dataset):
            raise ValueError(
                f"Prediction and ground truth dataset should have the same length. "
                f"Ground truth dataset has {len(resultset.ground_truth_dataset)} items, "
                f"prediction dataset has {len(resultset.prediction_dataset)} items"
            )
        resultset_labels = set(resultset.prediction_dataset.get_labels() + resultset.ground_truth_dataset.get_labels())
        model_labels = set(resultset.model.configuration.get_label_schema().get_labels(include_empty=False))
        labels = sorted(resultset_labels.intersection(model_labels))
        hard_predictions = []
        hard_references = []
        for prediction_item, reference_item in zip(
            list(resultset.prediction_dataset), list(resultset.ground_truth_dataset)
        ):
            hard_predictions.append(mask_from_dataset_item(prediction_item, labels))
            hard_references.append(mask_from_dataset_item(reference_item, labels))

        all_intersection, all_cardinality = get_intersections_and_cardinalities(
            hard_references, hard_predictions, labels
        )

        return cls.compute_dice_using_intersection_and_cardinality(all_intersection, all_cardinality, average)

    @classmethod
    def compute_dice_using_intersection_and_cardinality(
        cls,
        all_intersection: Dict[Optional[LabelEntity], int],
        all_cardinality: Dict[Optional[LabelEntity], int],
        average: MetricAverageMethod,
    ) -> Tuple[ScoreMetric, Dict[LabelEntity, ScoreMetric]]:
        """Computes dice score using intersection and cardinality dictionaries.

        Both dictionaries must contain the same set of keys.
        Dice score is computed by: 2 * intersection / cardinality

        Args:
            average: Averaging method to use
            all_intersection: collection of intersections per label
            all_cardinality: collection of cardinality per label

        Returns:
            A tuple containing the overall DICE score, and per label
            DICE score

        Raises:
            KeyError: if the keys in intersection and cardinality do not
                match
            KeyError: if the key `None` is not present in either
                all_intersection or all_cardinality
            ValueError: if the intersection for a certain key is larger
                than its corresponding cardinality
        """
        dice_per_label: Dict[LabelEntity, ScoreMetric] = {}

        for label, intersection in all_intersection.items():
            cardinality = all_cardinality[label]
            dice_score = cls.__compute_single_dice_score_using_intersection_and_cardinality(intersection, cardinality)

            # If label is None, then the dice score corresponds to the overall dice score
            # rather than a per-label dice score.
            # This score is calculated last because it can depend on the values in dice_per_label
            if label is not None:
                dice_per_label[label] = ScoreMetric(value=dice_score, name=label.name)

        # Set overall_dice to 0 in case the score cannot be computed
        overall_dice = ScoreMetric(value=0.0, name="Dice Average")
        if len(dice_per_label) == 0:  # dataset consists of background pixels only
            pass  # Use the default value of 0
        elif average == MetricAverageMethod.MICRO:
            overall_cardinality = all_cardinality[None]
            overall_intersection = all_intersection[None]
            dice_score = cls.__compute_single_dice_score_using_intersection_and_cardinality(
                overall_intersection, overall_cardinality
            )
            overall_dice = ScoreMetric(value=dice_score, name="Dice Average")
        elif average == MetricAverageMethod.MACRO:
            scores = [item.value for item in dice_per_label.values()]
            macro_average_score = sum(scores) / len(scores)
            overall_dice = ScoreMetric(value=macro_average_score, name="Dice Average")

        return overall_dice, dice_per_label

    @staticmethod
    def __compute_single_dice_score_using_intersection_and_cardinality(intersection: int, cardinality: int):
        """Computes a single dice score using intersection and cardinality.

        Dice score is computed by: 2 * intersection / cardinality

        Raises:
            ValueError: If intersection is larger than cardinality
        """
        if intersection > cardinality:
            raise ValueError("intersection cannot be larger than cardinality")
        if cardinality == 0 and intersection == 0:
            dice_score = 0.0
        else:
            dice_score = float(2 * intersection / cardinality)
        return dice_score



"""Unit Tests for the MPA Dataset Pipelines Transforms Augments."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


import random

import numpy as np
import pytest
from PIL import Image

from otx.mpa.modules.datasets.pipelines.transforms.augments import CythonAugments
from otx.mpa.modules.datasets.pipelines.transforms.augmix import (
    AugMixAugment,
    OpsFabric,
)


@pytest.fixture
def ops_fabric() -> OpsFabric:
    return OpsFabric("Rotate", 5, {"img_mean": 128})


@pytest.mark.xfail(reason="random may not return the same value on different machines.")
class TestOpsFabric:
    def test_init(self, ops_fabric: OpsFabric) -> None:
        """Test OpsFabric initialization."""
        assert ops_fabric.prob == 1.0
        assert ops_fabric.hparams == {"img_mean": 128}
        assert ops_fabric.aug_kwargs == {
            "fillcolor": 128,
            "resample": (Image.BILINEAR, Image.BICUBIC),
        }
        assert ops_fabric.magnitude == 5
        assert ops_fabric.magnitude_std == float("inf")
        assert ops_fabric.level_fn == ops_fabric._rotate_level_to_arg
        assert ops_fabric.aug_fn == CythonAugments.rotate

    def test_randomly_negate(self) -> None:
        """Test randomly_negate function."""
        random.seed(1234)
        assert OpsFabric.randomly_negate(5) == -5
        assert OpsFabric.randomly_negate(5) == 5
        assert OpsFabric.randomly_negate(5) == -5

    def test_rotate_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test rotate_level_to_arg function."""
        assert ops_fabric._rotate_level_to_arg(0, ops_fabric.hparams) == (0,)
        assert ops_fabric._rotate_level_to_arg(5, ops_fabric.hparams) == (5 / 10 * 30,)

    def test_enhance_increasing_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test enhance_increasing_level_to_arg function."""
        assert ops_fabric._enhance_increasing_level_to_arg(0, ops_fabric.hparams) == (1.0,)
        assert ops_fabric._enhance_increasing_level_to_arg(5, ops_fabric.hparams) == (1.0 + 5 / 10 * 0.9,)

    def test_shear_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test shear_level_to_arg function."""
        assert ops_fabric._shear_level_to_arg(0, ops_fabric.hparams) == (0,)
        assert ops_fabric._shear_level_to_arg(5, ops_fabric.hparams) == (5 / 10 * 0.3,)

    def test_translate_rel_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test translate_rel_level_to_arg function."""
        assert ops_fabric._translate_rel_level_to_arg(0, ops_fabric.hparams) == (0,)
        assert ops_fabric._translate_rel_level_to_arg(5, {"translate_pct": 0.5}) == (5 / 10 * 0.5,)

    def test_posterize_increasing_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test posterize_increasing_level_to_arg function."""
        assert ops_fabric._posterize_increasing_level_to_arg(0, ops_fabric.hparams) == (4,)
        assert ops_fabric._posterize_increasing_level_to_arg(5, ops_fabric.hparams) == (4 - int(5 / 10 * 4),)

    def test_solarize_increasing_level_to_arg(self, ops_fabric: OpsFabric) -> None:
        """Test solarize_increasing_level_to_arg function."""
        assert ops_fabric._solarize_increasing_level_to_arg(0, ops_fabric.hparams) == (0,)
        assert ops_fabric._solarize_increasing_level_to_arg(5, ops_fabric.hparams) == (256 - int(5 / 10 * 256),)

    def test_call(self, ops_fabric: OpsFabric) -> None:
        """Test __call__ function."""
        img = Image.new("RGB", (256, 256))
        transformed_img = ops_fabric(img)
        assert transformed_img != img  # make sure the image was actually transformed


class TestAugMixAugment:
    def test_init(self) -> None:
        """Test AugMixAugment initialization."""
        aug_mix_augment = AugMixAugment(config_str="augmix-m5-w3")
        assert isinstance(aug_mix_augment, AugMixAugment)
        assert len(aug_mix_augment.ops) > 0

    def test_apply_basic(self) -> None:
        """Test _apply_basic function."""
        aug_mix_augment = AugMixAugment(config_str="augmix-m5-w3")

        img = Image.new("RGB", (224, 224), color=(255, 0, 0))
        mixing_weights = np.float32(np.random.dirichlet([aug_mix_augment.alpha] * aug_mix_augment.width))
        m = np.float32(np.random.beta(aug_mix_augment.alpha, aug_mix_augment.alpha))

        mixed_img = aug_mix_augment._apply_basic(img, mixing_weights, m)
        assert isinstance(mixed_img, Image.Image)

    def test_augmix_ops(self) -> None:
        """Test augmix_ops function."""
        aug_mix_augment = AugMixAugment(config_str="augmix-m5-w3")
        assert len(aug_mix_augment.ops) > 0
        assert isinstance(aug_mix_augment.alpha, float)
        assert isinstance(aug_mix_augment.width, int)
        assert isinstance(aug_mix_augment.depth, int)

    def test_call(self) -> None:
        """Test __call__ method."""
        aug_mix_augment = AugMixAugment(config_str="augmix-m5-w3")
        data = {"img": np.random.randint(0, 255, size=(224, 224, 3)).astype(np.uint8)}
        results = aug_mix_augment(data)
        assert "augmix" in results
        assert isinstance(results["img"], Image.Image)



"""Base Self-SL dataset."""

__resize_target_size = 224
__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)

__train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations"),
    dict(
        type="TwoCropTransform",
        view0=[
            dict(type="RandomResizedCrop", size=__resize_target_size),
            dict(type="RandomFlip", prob=0.5, direction="horizontal"),
            dict(
                type="ProbCompose",
                transforms=[dict(type="ColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1)],
                probs=[0.8],
            ),
            dict(type="RandomGrayscale", p=0.2),
            dict(type="GaussianBlur", kernel_size=23),
            dict(type="Normalize", **__img_norm_cfg),
        ],
        view1=[
            dict(type="RandomResizedCrop", size=__resize_target_size),
            dict(type="RandomFlip", prob=0.5, direction="horizontal"),
            dict(
                type="ProbCompose",
                transforms=[dict(type="ColorJitter", brightness=0.4, contrast=0.4, saturation=0.2, hue=0.1)],
                probs=[0.8],
            ),
            dict(type="RandomGrayscale", p=0.2),
            dict(type="ProbCompose", transforms=[dict(type="GaussianBlur", kernel_size=23)], probs=[0.1]),
            dict(type="ProbCompose", transforms=[dict(type="Solarization", threshold=128)], probs=[0.2]),
            dict(type="Normalize", **__img_norm_cfg),
        ],
    ),
    dict(type="Collect", keys=["img", "gt_semantic_seg"]),
]

data = dict(samples_per_gpu=16, workers_per_gpu=2, train=dict(type="MPASegDataset", pipeline=__train_pipeline))
